//
//  Copyright 2012-2013, Richard Gorton      All Rights Reserved
//
/// \file
/// Definitions for characteristics of path and strata cycles - the CycleCharacteristics class
///
/// \author
/// Richard Gorton
///

#ifndef _CYCLES_H_
#define _CYCLES_H_

#include <map>
#include <vector>
#include <algorithm>

#include "pprof.h"

/// CycleCharacteristics - data pertinent to all types of paths being collected, and relevant
///                       to each individual path
class CycleCharacteristics {
    public:
        CycleCharacteristics() :
            m_cycleID(s_nullCycleID),
            m_numElements(0),   // This may be moot for stl based containers
            m_tripCount(0),
            m_pathHashVal(0)
        {}

        ~CycleCharacteristics() {}

        /// Get values
        const CycleID_T  CycleID()      { return m_cycleID; }
        const PathLen_T NumElements()   { return m_numElements; }
        const TripCount_T  TripCount()  { return m_tripCount; }
        const uint64_t  HashVal()       { return m_pathHashVal; }

        /// Set values
        void SetCycleID(const CycleID_T path)       { m_cycleID = path; }
        void SetNumElements(const PathLen_T num)    { m_numElements = num; }
        void SetTripCount(const TripCount_T num)    { m_tripCount = num; }
        void SetHashVal(const uint64_t hash)        { m_pathHashVal = hash; }

        /// Modify values
        void ModifyHashVal(const uint64_t adjust)   { m_pathHashVal += adjust; }
        void IncTripCount(const TripCount_T adjust) { m_tripCount += adjust; }

        /// want a virtual RawPathVector, but definitions are class specific
        /// The ExactMatch method would seem to be a template thingie, except
        /// that "Exact" may be fuzzy while experimenting (at least for DataPaths)

        /// Keep these private - Meyers, item 22.  Always encapsulate
    private:
        CycleID_T       m_cycleID;
        PathLen_T       m_numElements;
        TripCount_T     m_tripCount;
        uint64_t        m_pathHashVal;
};

/// InsnPath - characteristics specific to ipaths (instruction paths)
class InsnPath : public CycleCharacteristics {
    public:

        InsnPath() :
            m_numInsns(0)
        {
            m_pathVec.clear();
            // is a m_pathVec.reserve() needed?
        }

        ~InsnPath() {}

        /// Get a copy of the raw vector
        const InsnPath_T& RawPathVector()               { return this->m_pathVec; }

        /// The size of the path
        const PathLen_T size()                          { return m_pathVec.size(); }

        /// Get number of Insns
        const PathLen_T PathInsnLen()                   { return m_numInsns; }

        /// Search for an exact match of an element in the current vector
        /// \param[in]  InsnPathElt_T - the Ipath element to search for
        /// \returns    true if found, false otherwise
        bool ExactMatch(const Address_T BbAddr)
        {
            InsnPath_T::const_iterator  where;

            // Vectors do not have a find
            where = find(m_pathVec.begin(), m_pathVec.end(), BbAddr);
            if (where == m_pathVec.end())
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// Add the element to the path vector
        /// \param[in]  InsnPathElt_T - Ipath element to append
// ? can this be a template?
// presume we don't fail due to memory...
        void ExtendIPath(const Address_T& BbAddr, const uint32_t& insnCount)
                {
                m_pathVec.push_back(BbAddr);
                m_numInsns += insnCount;
                // Meyers item27 - reinterpret_cast because it is really low level
                ModifyHashVal(reinterpret_cast<uint64_t>(BbAddr));
                }

        /// Reset the immediately current path.  Call this when further
        /// downstream processing has been completed.
        void ResetPath()
                {
                m_pathVec.clear();
                m_numInsns = 0;
                SetHashVal(0);
                SetTripCount(0);
                }

    private:
        InsnPath_T      m_pathVec;
        PathLen_T       m_numInsns; // Number of instructions in this Path;
        // what about micro-stalls etc as the heat/number of cycles for a path?
        // Hmmm.  How to track inconsistent behavior - periodic dcache misses
        // on a path?  Flip it around: what about paths which get consistent
        // dcache misses?  Can the code sequences be rewritten to avoid them?
};

/// DataPath - characteristics specific to dpaths (data address paths)
class DataPath : public CycleCharacteristics {
    public:
        DataPath() :
            m_numReads(0),
            m_numWrites(0),
            m_numAtomicUpdates(0),
            m_numPrefetches(0),
            m_numSpanPageReads(0),
            m_numSpanPageWrites(0)
        {
            m_pathVec.clear();
        }

        ~DataPath() {}

        /// Get a copy of the raw vector
        const DataCycle_T& RawPathVector()      { return this->m_pathVec; }

        /// The size of the path
        const PathLen_T size()                  { return m_pathVec.size(); }

        /// Search for an exact match of an element in the current vector
        /// May be a fuzzy function depending upon the experiment being run;
        /// consequently implemented in the .cpp file
        /// \param[in]  DataPathElt_T - the Dpath element to search for
        /// \returns    true if found, false otherwise
        bool ExactMatch(const DataPathElt_T& theItem);

        // Extend the path is a triple: address, width, type
        /// Add the element to the path vector
        /// \param[in]  DataPathElt_T - Ipath element to append
        void ExtendCycle(const DataPathElt_T& theItem);

        /// Reset the path (perhaps lighter weight than full destruction/construction)
        void ResetPath()
        {
            m_pathVec.clear();
            m_numReads          = 0;
            m_numWrites         = 0;
            m_numAtomicUpdates  = 0;
            m_numPrefetches     = 0;
            m_numSpanPageReads  = 0;
            m_numSpanPageWrites = 0;
        }

        /// Class specific values

        /// the Get functions
        const uint64_t      NumReads()          { return m_numReads; }
        const uint64_t      NumWrites()         { return m_numWrites; }
        const uint64_t      NumAtomicUpdates()  { return m_numAtomicUpdates; }
        const uint64_t      NumPrefetches()     { return m_numPrefetches; }
        const uint64_t      NumSpanPageReads()  { return m_numSpanPageReads; }
        const uint64_t      NumSpanPageWrites() { return m_numSpanPageWrites; }

        /// Setting values
        void SetNumReads(uint64_t num)          { m_numReads = num; }
        void SetNumWrites(uint64_t num)         { m_numWrites = num; }
        void SetNumAtomicUpdates(uint64_t num)  { m_numAtomicUpdates = num; }
        void SetNumPrefetches(uint64_t num)     { m_numPrefetches = num; }
        void SetNumSpanPageReads(uint64_t num)  { m_numSpanPageReads = num; }
        void SetNumSpanPageWrites(uint64_t num) { m_numSpanPageWrites = num; }

    private:
        DataCycle_T     m_pathVec;
        /// speculative: Do we need to know the number of Bb's involved?
        /// the actual Bb IDs or iPath AddressPaths_T
        uint64_t        m_numReads;
        uint64_t        m_numWrites;
        uint64_t        m_numAtomicUpdates;
        uint64_t        m_numPrefetches;
        uint64_t        m_numSpanPageReads;
        uint64_t        m_numSpanPageWrites;
};

/// StrataCycle - characteristics specific to strata layers (cycles of repeated paths)
class StrataCycle : public CycleCharacteristics {
    public:

        StrataCycle()
        {
            m_pathVec.clear();
        }

        ~StrataCycle() {}

        /// Get a copy of the raw vector
        const StrataCycle_T& RawPathVector()            { return this->m_pathVec; }

        /// Assign a value to the raw vector
        void SetRawPathVector(const StrataCycle_T& vec) { m_pathVec = vec; }

        /// The size of the path
        const SLen_T size()                             { return m_pathVec.size(); }

        /// Search for an exact match of an element in the current vector
        /// Needs to be in the .cpp file because we need global overloads of
        /// the "==" operator
        /// \param[in]  InsnPathElt_T - the Ipath element to search for
        /// \returns    true if found, false otherwise
        bool ExactMatch(const StratumElt_T searchForElt);

        /// Add the element to the vector
        void ExtendCycle(const StratumElt_T& newElt)
                {
                m_pathVec.push_back(newElt);
                ModifyHashVal(newElt.Elt_ID);
                ModifyHashVal(newElt.Elt_Trips);
                }

        /// Reset the immediately current path.  Call this when further
        /// downstream processing has been completed.
        void ResetCycle()
                {
                m_pathVec.clear();
                SetHashVal(0);
                SetTripCount(0);
                }

    private:
        StrataCycle_T   m_pathVec;
};
#endif  // _CYCLES_H_
