//
//  Copyright 2012-2013, Richard Gorton      All Rights Reserved
//
/// \file
/// Implementation of Istream Path Profiler (Instruction)
/// This file contains the execution-time content for the Data Access path profiler
///
/// \author
/// Richard Gorton          rcgorton@verizon.net
///             
/// Data address path analyzer code.
/// The general hypothesis is that much like istream addresses, data access
/// addresses can be compactly represented as repeated paths.
///                      
/// The catch is that for data accesses, the pertinent information is all of:
///     base address     
///     read/write/[both] behavior
///     width of access  
/// Instruction stream paths are read-only, and while it is possible to do so,
/// 'normal' code does not enter the middle of another instruction.
/// Skip/ignore this possibility, as it would be either contrived code, or
/// malware.  Fixed instruction width architectures typically require instructions
/// to start on 0mod<size> boundaries, which makes it impossible to have valid
/// code enter the middle of an instruction.
///         
/// However, data accesses can be variable width.  A write access 1-byte wide to
/// address A+3 is going to affect 4-byte reads from adddress A.
/// And in some architectures, the underlying/backing hardware may have a wider
/// granularity of access, with some sort of merge layer to do some level of
/// resolving.  
///                          
/// Additionally, the actions instructions can have:
///     Read                 
///     Write
///     Both (probably infrequent)
///     Unknown: is it useful to differentiate atomic accesses as part of the
///              definition of a cycle?  Atomics are key for multi-threaded ops
///              and/or I/O, but these experiments are going to be fed single-threaded,
///              user-space code.
///             
/// Thus, the experiments (EXPERIMENT_TYPE): construct our data address paths in a various ways,
/// by varying the definition of the completion of a path.
/// 1) A match of the address and action (read-or-write-or-both)
/// 2) A match of the address, action and width (exact match on all)
/// 3) A match on the action, and any of the bytes referenced by in the range
///    [address, address+width)
/// This primarily affects the ExactMatch function
///

#include <limits>
#include <stdint.h>
#include <map>
#include <algorithm>
#include <cstring>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <zlib.h>

#include "util.h"
#include "pprof.h"
#include "CmdOptions.h"
#include "PathProfiler.h"
#include "Cycles.h"
#include "pstats.h"
#include "Logger.h"
#include "dbio.h"

/// C interfaces to the execution-time data collection for DPath profiling
extern "C" {
    void DPathInit(void*);
    void ConstructDPaths(void* memAddr, int accessSize, uint32_t accessMode);
    void DPathFinish(int, void*);
}

static DataPath         s_prevPathCache;
static DataPath*        s_currentPath = NULL;
static DCycleStats*     s_DStats = NULL;

// the next one available to be assigned (these are unique)
static CycleID_T        s_nextCycleID = s_nullCycleID;

// s_rawIdx is the offset into the shared mem data area
static ShmArea_T    s_localSharedMem;
static size_t       s_rawIdx = 0;

// Only meaningful if we dump the pathid/tripcounts to a file
static int          s_dumpFD = -1;

/// The is the primary storage mechanism - a map: Path --> Info
static DataCycleToInfo_T*     s_dataCycleToInfo = NULL;

static int  s_pageSize;
static int  s_pageSizeMask;
/// Initialization of execution-time items
void DPathInit(void*)
{
    // This function is set up via a PIN_AddApplicationStartFunction()
    // callback which gets called once every image load.  Right now, this is now
    // how the istream gets instrumented, but this may change.

    static bool InitDone = false;

    if (InitDone)
    {
        return;
    }

    s_pageSize = getpagesize();
    s_pageSizeMask = s_pageSize - 1;

    if (NULL == s_currentPath)
    {
        s_currentPath = new DataPath();
    }
    if (NULL == s_currentPath)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_currentPath\n");
    }

    if (NULL == s_DStats)
    {
        s_DStats = new DCycleStats();
    }
    if (NULL == s_DStats)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_DStats\n");
    }

    if ((PROFILEMODE_SAVEPATHS == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        // Construct the map to store the paths
        // IPath --> ID
        if (NULL == s_dataCycleToInfo)
        {
            s_dataCycleToInfo = new DataCycleToInfo_T();
        }
        if (NULL == s_dataCycleToInfo)
        {
            LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_dataCycleToInfo\n");
        }
    }

    if (s_nullCycleID == s_nextCycleID)
    {
        s_nextCycleID = s_minCycleID;
    }

    // Initialization of the local shared memory copy.
    s_localSharedMem.dataHdr.version            = s_shm_verno;
    s_localSharedMem.dataHdr.operation          = OP_UNINITIALIZED;
    s_localSharedMem.dataHdr.collectionSet      = g_PathProfiler->CollectionSet() | g_PathProfiler->ProfileMode();
    s_localSharedMem.dataHdr.shm_flags          = SHMFLAG_UNINITALIZED;
    s_localSharedMem.dataHdr.payloadSize        = 0;
    s_localSharedMem.dataHdr.startInsnEpoch     = 0;
    s_localSharedMem.dataHdr.startPathEpoch     = 0;
    s_localSharedMem.dataHdr.startStrataEpoch   = 0;
    strcpy (s_localSharedMem.dataHdr.dataFileBaseName, g_PathProfiler->ProfilerBaseName().c_str());

    if (g_CmdOptions->DumpPathsToFile())
    {
        // Instead of writing to shared memory, we will end up writing the shared
        // memory buffers to a file
        int     oFlags = O_RDWR|O_CREAT|O_EXCL|O_TRUNC;
        mode_t  permissions = 0666;
        unlink(g_PathProfiler->StrataBuilderMapFile().c_str());
        s_dumpFD = open(g_PathProfiler->StrataBuilderMapFile().c_str(), oFlags, permissions);
        if (-1 == s_dumpFD)
        {
            LogAndAbort(__FUNCTION__, "Unable to create the file to dump the pathid/tripcounts into\n");
        }
    }

    InitDone = true;
}

/// Write the pathinfo db content out
static void WritePathInfoDB()
{
    CycleID_T   sanityCheck;
    PureDBWrite pathInfoDB;

    pathInfoDB.Open(g_PathProfiler->ProfilerPathInfoDBName());

    // Header
    pathInfoDB.Write(s_dbVerNumKey, s_dbVerNumValue);

    // Common statistics.  See dbio.h for the conventions used
    // feels like some sort of table would be more appropriate
    pathInfoDB.Write("PathDictionary:1", s_dbStatsTotalCycles);
    pathInfoDB.Write(s_dbStatsTotalCycles, s_DStats->TotalCycles());

    pathInfoDB.Write("PathDictionary:2", s_dbStatsUniqueCycles);
    pathInfoDB.Write(s_dbStatsUniqueCycles, s_DStats->UniqueCycles());

    pathInfoDB.Write("PathDictionary:3", s_dbStatsMaxCycleLength);
    pathInfoDB.Write(s_dbStatsMaxCycleLength, s_DStats->MaxCycleLength());

    pathInfoDB.Write("PathDictionary:4", s_dbStatsMaxPathHeat);
    pathInfoDB.Write(s_dbStatsMaxPathHeat, s_DStats->MaxPathHeat());

    pathInfoDB.Write("PathDictionary:5", s_dbStatsMaxTripCount);
    pathInfoDB.Write(s_dbStatsMaxTripCount, s_DStats->MaxTripCount());

    pathInfoDB.Write("PathDictionary:6", s_dbStatsMaxImmediateTripCount);
    pathInfoDB.Write(s_dbStatsMaxImmediateTripCount, s_DStats->MaxImmediateTripCount());

    pathInfoDB.Write("PathDictionary:7", s_dbStatsCachedHits);
    pathInfoDB.Write(s_dbStatsCachedHits, s_DStats->CachedHits());

    // DataCycle specific statistics
    pathInfoDB.Write("PathDictionary:8", s_dbStatsTotalReads);
    pathInfoDB.Write(s_dbStatsTotalReads, s_DStats->TotalReads());

    pathInfoDB.Write("PathDictionary:9", s_dbStatsTotalWrites);
    pathInfoDB.Write(s_dbStatsTotalWrites, s_DStats->TotalWrites());

    pathInfoDB.Write("PathDictionary:10", s_dbStatsTotalAtomics);
    pathInfoDB.Write(s_dbStatsTotalAtomics, s_DStats->TotalAtomics());

    pathInfoDB.Write("PathDictionary:11", s_dbStatsTotalPrefetches);
    pathInfoDB.Write(s_dbStatsTotalPrefetches, s_DStats->TotalPrefetches());

    pathInfoDB.Write("PathDictionary:12", s_dbStatsTotalSpanPageReads);
    pathInfoDB.Write(s_dbStatsTotalSpanPageReads, s_DStats->TotalSpanPageReads());

    pathInfoDB.Write("PathDictionary:13", s_dbStatsTotalSpanPageWrites);
    pathInfoDB.Write(s_dbStatsTotalSpanPageWrites, s_DStats->TotalSpanPageWrites());

    pathInfoDB.Write("PathDictionary:14", s_dbStatsMaxReads);
    pathInfoDB.Write(s_dbStatsMaxReads, s_DStats->MaxReads());

    pathInfoDB.Write("PathDictionary:15", s_dbStatsMaxWrites);
    pathInfoDB.Write(s_dbStatsMaxWrites, s_DStats->MaxWrites());

    pathInfoDB.Write("PathDictionary:16", s_dbStatsMaxAtomics);
    pathInfoDB.Write(s_dbStatsMaxAtomics, s_DStats->MaxAtomics());

    pathInfoDB.Write("PathDictionary:17", s_dbStatsMaxPrefetches);
    pathInfoDB.Write(s_dbStatsMaxPrefetches, s_DStats->MaxPrefetches());

    pathInfoDB.Write("PathDictionary:18", s_dbStatsMaxSpanPageReads);
    pathInfoDB.Write(s_dbStatsMaxSpanPageReads, s_DStats->MaxSpanPageReads());

    pathInfoDB.Write("PathDictionary:19", s_dbStatsMaxSpanPageWrites);
    pathInfoDB.Write(s_dbStatsMaxSpanPageWrites, s_DStats->MaxSpanPageWrites());

    pathInfoDB.Write("PathDictionary:20", s_dbStatsMaxAccessWidth);
    pathInfoDB.Write(s_dbStatsMaxAccessWidth, s_DStats->MaxAccessWidth());

    uint64_t    numEntries = 20;
    pathInfoDB.Write("PathDictionary", numEntries);

    // And finally the contents of each unique path.
    // s_nullCycleID because we increment at the end of the loop
    sanityCheck = s_nullCycleID;

    for (DataCycleToInfo_T::const_iterator c_iter = s_dataCycleToInfo->begin(); c_iter != s_dataCycleToInfo->end(); c_iter++)
    {
        std::string     theIDKey;
        std::string     theVectorKey;
        PathInfo_T      thisInfo;

        thisInfo = c_iter->second;
        theIDKey = s_dbPathInfoKey + ToString(thisInfo.cycleID);
        pathInfoDB.Write(theIDKey, thisInfo);

        DataCycle_T  AddrVec = c_iter->first;
        theVectorKey = s_dbDPathIDKey + ToString(thisInfo.cycleID);

        /// Note: 6.2.3 of "The C++ Standard Library" by Josuttis notes that elements of a vector
        /// can be expected to be in contiguous memory.  Thus, we merely have to set the call
        /// to Write() up carefully, and do not need to perform a new/copy/delete.
        pathInfoDB.Write(const_cast<char*>(theVectorKey.c_str()),
                         reinterpret_cast<char*>(&AddrVec[0]), 
                         AddrVec.size() * sizeof(AddrVec[0]));
        sanityCheck++;
    }

    if (s_DStats->UniqueCycles() != sanityCheck)
    {
        LogAndAbort(__FUNCTION__, "Consistency error: mismatch on number of unique paths vs. number in the map\n");
    }

    pathInfoDB.Close();
}

// Type 1: match on address/action
// Type 2: match on address/action/width
// Type 3: match on address/action, possible width overlap
//         on range [address, address+width)
#define EXPERIMENT_TYPE 1

/// Operator overloads to enable simple comparisons of the vectors based upon
/// an article on cplusplus.com: "compare vectors for equality"
/// I tried putting it inside the class definition for StrataCycle, but that fails
/// at compile time "error: ... must take exactly one argument"
/// there is a response that it needs to be overloaded at global scope.
/// putting it into a .h file resulted in errors about it being multiply defined.
/// Need an operator for "==" to enable:
///         if (searchForElt == *where)
/// This makes the ExactMatch code quite straitforward
bool operator == (const DataPathElt_T& p1, const DataPathElt_T& p2)
{
#if (EXPERIMENT_TYPE == 1)
    //  1) A match of the address and action (read-or-write-or-both)
    if ((p1.address == p2.address) && (p1.mode == p2.mode))
    {
        return true;
    }
    return false;
#elif (EXPERIMENT_TYPE == 2)
    //  2) A match of the address, action and width (exact match on all)
    if ((p1.address == p2.address) && (p1.width == p2.width) && (p1.mode == p2.mode))
    {
        return true;
    }
    return false;
#elif (EXPERIMENT_TYPE == 3)
    //  3) A match on the action, and any of the bytes referenced by in the range
    //     [address, address+width)
    Address_T   vecBase;
    Address_T   vecEnd;
    Address_T   theBase;
    Address_T   theEnd;

    vecBase = p1.address;
    theBase = p2.address;

    // Which cast to use?
    vecEnd = vecBase + p1.width;
    theEnd = theBase + p2.width;

    if (p1.mode != p2.mode)
    {
        return false;
    }
    if ((vecEnd < theBase)
        || (vecBase > theEnd))
    {
        return false;
    }
    return true;
#else
#error "EXPERIMENT_TYPE undefined or has an invalid value - needs to be (1..3)"
#endif
}

/// this overload is needed for map insertion:
///         (*s_StrataCycletoInfo)[findMe->RawPathVector()] = <value>;
bool operator < (const DataPathElt_T& p1, const DataPathElt_T& p2)
{
    if (p1.address < p2.address)
    {
        return true;
    }
    if ((p1.address == p2.address) && (p1.width < p2.width))
    {
        return true;
    }
    // hmm. mode bits as a comparison?
    if ((p1.address == p2.address) && (p1.width == p2.width)
        && (p1.mode < p2.mode))
    {
        return true;
    }
    return false;
}

bool DataPath::ExactMatch(const DataPathElt_T& theItem)
{
    DataCycle_T::const_iterator     where;

    for (where = m_pathVec.begin(); where != m_pathVec.end(); where++)
    {
        if (*where == theItem)
        {
            return true;
        }
    }
    return false;
}

void DataPath::ExtendCycle(const DataPathElt_T& theItem)
{
    // Meyers item27 - reinterpret_cast because it is really low level
    ModifyHashVal(reinterpret_cast<uint64_t>(theItem.address));
    ModifyHashVal(theItem.width);
    ModifyHashVal(theItem.mode);

    m_pathVec.push_back(theItem);
    if (s_accessBitsRead == (theItem.mode & s_accessBitsRead))
    {
        m_numReads++;
    }
    if (s_accessBitsWrite == (theItem.mode & s_accessBitsWrite))
    {
        m_numWrites++;
    }
    if (s_accessBitsAtomicUpdate == (theItem.mode & s_accessBitsAtomicUpdate))
    {
        m_numAtomicUpdates++;
    }
    if (s_accessBitsPrefetch == (theItem.mode & s_accessBitsPrefetch))
    {
        m_numPrefetches++;
    }

    // page boundaries
    uint64_t startPageAddr  = reinterpret_cast<uint64_t>(theItem.address);
    uint64_t finishPageAddr = startPageAddr + theItem.width;
    uint64_t startPageBase  = startPageAddr & ~s_pageSizeMask;
    uint64_t finishPageBase = finishPageAddr & ~s_pageSizeMask;
    if (startPageBase != finishPageBase)
    {
        if (s_accessBitsRead == (theItem.mode & s_accessBitsRead))
        {
            m_numSpanPageReads++;
        }
        if (s_accessBitsWrite == (theItem.mode & s_accessBitsWrite))
        {
            m_numSpanPageWrites++;
        }
    }
}

/// A utility function to do the legwork of looking for a path in the maps, or creating
/// a new one.  Most of the time this is going to be passed the cached version, but for
/// the final flush, we also need to process the currently-being-built path.
/// \param[in]  the InstructionPath to be looked for
/// \param[out] the cycleID (a new one if it does not already exist)
/// \param[out] the paths TripCount
void FindOrCreatePath(DataPath* findMe, CycleID_T* temporalID, TripCount_T* temporalCount)
{
    CycleID_T                   matchedID = 0;
    TripCount_T                 numTrips;
    DataCycleToInfo_T::iterator     iter;

    numTrips = findMe->TripCount();

    iter = s_dataCycleToInfo->find(findMe->RawPathVector());
    if (iter != s_dataCycleToInfo->end())
    {
        // path pre-exists.
        matchedID = iter->second.cycleID;

        iter->second.tripCount += numTrips;

        if (s_DStats->MaxTripCount() < iter->second.tripCount)
        {
            s_DStats->SetMaxTripCount(iter->second.tripCount);
        }
    }   // was in DataCycleToInfo map
    else
    {
        // New path.  Create the info and both of the addressvector/ID maps
        PathInfo_T  pInfo;

        matchedID = s_nextCycleID;

        pInfo.cycleID = s_nextCycleID;
        pInfo.tripCount = numTrips;
        pInfo.pathHeat = findMe->size();
        pInfo.pathLen = findMe->size();

        (*s_dataCycleToInfo)[findMe->RawPathVector()] = pInfo;

        // Need to allocate a new Cycleid.
        s_nextCycleID++;
        s_DStats->IncUniqueCycles(1);
        s_DStats->UniqueLengthHistogram()[findMe->size()]++;
    }
    *temporalID = matchedID;
    *temporalCount = numTrips;
}

/// Construct the paths and process them - this performs the bulk of the execution-time work
/// Called from the PIN instrumentation code.
/// \param[in]      access address
/// \param[in]      access size in bytes
/// \param[in]      access mode.  Comes in as a uint32_t
void ConstructDPaths(void* memAddr, int accessSize, uint32_t accessMode)
{
    bool            quickMatch = false;
    CycleID_T       temporalID = s_nullCycleID;
    TripCount_T     temporalCount = 0;
    DataPathElt_T   pathElt;

    pathElt.address = memAddr;
    pathElt.width = accessSize;
    pathElt.mode = accessMode;

    s_DStats->IncTotalAccesses(1);
    if (s_DStats->MaxAccessWidth() < pathElt.width)
    {
        s_DStats->SetMaxAccessWidth(pathElt.width);
    }
    s_DStats->AccessHistogram()[pathElt.width]++;

    if (!s_currentPath->ExactMatch(pathElt))
    {
        // add to the end of the potential cycle
        s_currentPath->ExtendCycle(pathElt);
        return;
    }

    // We found a match, it means we had a cycle; the level of processing
    // depends a lot on what the invocation settings were.

    // Always collect length information/statistics
    // Parent class
    s_DStats->IncTotalCycles(1);
    if (s_DStats->MaxCycleLength() < s_currentPath->size())
    {
        s_DStats->SetMaxCycleLength(s_currentPath->size());
    }
    s_DStats->DynamicLengthHistogram()[s_currentPath->size()]++;

    // For the simplistic approach, the Heat of a path is the length in instructions
    // A more interesting approach would be to calculate the number of cycles required
    // as computed on a static basis (that is, ignoring things like memory conflicts/stalls)
    PathHeat_T currPathHeat = s_currentPath->size();
    if (s_DStats->MaxPathHeat() > currPathHeat)
    {
        s_DStats->SetMaxPathHeat(currPathHeat);
    }

    // DataPath specific statistics
    s_DStats->IncTotalReads(s_currentPath->NumReads());
    if (s_DStats->MaxReads() < s_currentPath->NumReads())
    {
        s_DStats->SetMaxReads(s_currentPath->NumReads());
    }

    s_DStats->IncTotalWrites(s_currentPath->NumWrites());
    if (s_DStats->MaxWrites() < s_currentPath->NumWrites())
    {
        s_DStats->SetMaxWrites(s_currentPath->NumWrites());
    }

    // Atomics covered by read/write above
    s_DStats->IncTotalAtomics(s_currentPath->NumAtomicUpdates());
    if (s_DStats->MaxAtomics() < s_currentPath->NumAtomicUpdates())
    {
        s_DStats->SetMaxAtomics(s_currentPath->NumAtomicUpdates());
    }

    s_DStats->IncTotalPrefetches(s_currentPath->NumPrefetches());
    if (s_DStats->MaxPrefetches() < s_currentPath->NumPrefetches())
    {
        s_DStats->SetMaxPrefetches(s_currentPath->NumPrefetches());
    }

    // Page spanning
    s_DStats->IncTotalSpanPageReads(s_currentPath->NumSpanPageReads());
    if (s_DStats->MaxSpanPageReads() < s_currentPath->NumSpanPageReads())
    {
        s_DStats->SetMaxSpanPageReads(s_currentPath->NumSpanPageReads());
    }

    s_DStats->IncTotalSpanPageWrites(s_currentPath->NumSpanPageWrites());
    if (s_DStats->MaxSpanPageWrites() < s_currentPath->NumSpanPageWrites())
    {
        s_DStats->SetMaxSpanPageWrites(s_currentPath->NumSpanPageWrites());
    }

    // And class specific content which depends upon the profiling mode
    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        goto resetfornext;
    }


    // For PROFILEMODE_SAVEPATHS, PROFILEMODE_STRATALENGTH, and PROFILEMODE_SAVESTRATA,
    // we need to perform further processing

    // Performance kicker:
    //  - while building the current path, keep a simple checksum as a quick check
    //    to possibly avoid the overhead of the find in the map; keep a copy of 
    //    the immediately previous vector/ID, so that if the checksum does match,
    //    we can try the immediate compare, and do a simple handoff.  If the
    //    checksum does not match, then yes we do have to do the find.  But tight
    //    loops are going to hit a lot.  Perhaps even a larger quickcache to
    //    catch a couple levels of nesting and/or simple control flow inside of
    //    loops in PROFILEMODE_SAVEPATHS (but not either of the STRATA collectors).

    if (s_prevPathCache.HashVal() == s_currentPath->HashVal())
    {
        if (s_prevPathCache.RawPathVector() == s_currentPath->RawPathVector())
        {
            quickMatch = true;
            s_DStats->IncCachedHits(1);
            s_prevPathCache.IncTripCount(1);
            if (s_DStats->MaxImmediateTripCount() < s_prevPathCache.TripCount())
            {
                s_DStats->SetMaxImmediateTripCount(s_prevPathCache.TripCount());
            }
        }
    }

    // Once the cache is being used, there are a number of cases to consider
    //  Pxn - Path <x>, trip #n, so PA1, PB1, PB2, PC1, etc.
    // Initial entry at program start
    //      PA1, PA2, PB1
    // Subsequent:
    //      PC1, PC2, PD1, PE1
    // Algorithm: if not in cache, find cache in the map
    //      if there, add in cache # trips
    //      if not, create with cache # trips

    // be sure not to create an entry for very first path
    if (!quickMatch && (0 < s_prevPathCache.size()))
    {
        FindOrCreatePath(&s_prevPathCache, &temporalID, &temporalCount);
    }

    if (!quickMatch)
    {
        s_prevPathCache = *s_currentPath;
        s_prevPathCache.SetTripCount(1);
    }

    // The temporal approach requires that we save a vector of CycleIDs.  The original
    // implementation took advantage of the fact that many paths are immediately repeated
    if ((PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        if (s_nullCycleID != temporalID)
        {
            // handoff through shared memory
            g_PathProfiler->PassToStrataServer(temporalID, temporalCount, false);
        }
    }

resetfornext: ;
    // Now that we have processed everything, reset the current path,
    // and start construction of a new path
    s_currentPath->ResetPath();
    s_currentPath->ExtendCycle(pathElt);
}

/// Pass a CycleID/tripCount to the Strata server process.
/// This happens via shared memory; we need to try to keep that memory as free from
/// contention on both sizes where possible.  Without being sure where potential bottlenecks
/// are, it is a reasonable first guess (due to the injection of the instrumentation code)
/// that the overhead incurred by the producer is larger than that of the consumer.
/// The original version kept a local copy of the buffer and built on that; when the
/// local copy got full, a memcpy into the shared memory area was done.  It might be
/// the case that a different set of semantics is better.  Analysis will be required
/// to determine that.  (aka To Be Determined later).
void DataPathProfiler::PassToStrataServer (CycleID_T id, TripCount_T trips, bool finalFlush)
{
    PublicStratumElt_T  elt;

    elt.cycleID = id;
    elt.tripCount = trips;

    // buffer is full, or we need to complete the processing and shut down
    if ((s_rawIdx + sizeof(elt) > s_ShmemBufSize) || finalFlush)
    {
        char    opFlags;

        opFlags = OP_UNINITIALIZED;
        if (0 == s_DStats->SharedMemBufferCount())
        {
            opFlags |= OP_OPEN_APPEND;
        }
        if (finalFlush) {
            opFlags |= OP_FLUSH_CLOSE;
        }

        s_DStats->IncShmBufferCount(1);

        // Build up a message for the server to analyze/compress
        s_localSharedMem.dataHdr.version        = s_shm_verno;
        s_localSharedMem.dataHdr.operation      = opFlags;
        s_localSharedMem.dataHdr.collectionSet  = g_PathProfiler->CollectionSet() | g_PathProfiler->ProfileMode();
        s_localSharedMem.dataHdr.payloadSize    = s_rawIdx;
        s_localSharedMem.dataHdr.startInsnEpoch = s_DStats->TotalAccesses();
        s_localSharedMem.dataHdr.startPathEpoch = s_DStats->TotalCycles();
        s_localSharedMem.dataHdr.startStrataEpoch = 0;

        if (g_CmdOptions->DumpPathsToFile())
        {
            // Simply mark this buffer as "filled" and write the entire area to file
            // So our physical file will be a multiple of ShmArea_T bytes
            s_localSharedMem.dataHdr.shm_flags = SHMFLAG_FILLED;
            write(s_dumpFD, &s_localSharedMem, sizeof(ShmArea_T));
        }
        else
        {
            // Send the message to the server, once server is available
            // Technique for debugging the strata server:
            //  Break on the MarkDataFilled statement in the path generator;
            //  In the strata server, set a break on the WaitUntilReadyToFill
            WaitUntilReadyToFill (g_PathProfiler->StrataSharedMem());
            memcpy ((void*) g_PathProfiler->StrataSharedMem(), (void*) &s_localSharedMem, sizeof(s_localSharedMem));
            MarkDataFilled (const_cast<ShmArea_T*>(g_PathProfiler->StrataSharedMem()));
        }

        // reset for further use
        s_rawIdx = 0;
    }

        // copy new info into buffer
    memcpy ((void*) &s_localSharedMem.rawData[s_rawIdx], (void*) &elt, sizeof(elt));
    s_rawIdx += sizeof(elt);

}


/// Flush the last path (and flush the strata as well if relevant)
/// In this case, we will need to flush the cached content, along with the
/// "current" content.
static void FlushLastPath()
{
    CycleID_T   cachedTemporalID = s_nullCycleID;
    CycleID_T   finalTemporalID = s_nullCycleID;
    TripCount_T cachedTemporalCount = 0;
    TripCount_T finalTemporalCount = 0;

    // Whatever partial path is there is now "complete"
    s_DStats->IncTotalCycles(1);
    if (s_DStats->MaxCycleLength() < s_currentPath->size())
    {
        s_DStats->SetMaxCycleLength(s_currentPath->size());
    }
    s_DStats->DynamicLengthHistogram()[s_currentPath->size()]++;

    // For the simplistic approach, the Heat of a path is the length in instructions
    // A more interesting approach would be to calculate the number of cycles required
    // as computed on a static basis (that is, ignoring things like memory conflicts/stalls)
    PathHeat_T currPathHeat = s_currentPath->size();
    if (s_DStats->MaxPathHeat() > currPathHeat)
    {
        s_DStats->SetMaxPathHeat(currPathHeat);
    }

    // DataPath statistics
    if (s_DStats->MaxReads() < s_currentPath->NumReads())
    {
        s_DStats->SetMaxReads(s_currentPath->NumReads());
    }

    if (s_DStats->MaxWrites() < s_currentPath->NumWrites())
    {
        s_DStats->SetMaxWrites(s_currentPath->NumWrites());
    }

    if (s_DStats->MaxAtomics() < s_currentPath->NumAtomicUpdates())
    {
        s_DStats->SetMaxAtomics(s_currentPath->NumAtomicUpdates());
    }

    if (s_DStats->MaxPrefetches() < s_currentPath->NumPrefetches())
    {
        s_DStats->SetMaxPrefetches(s_currentPath->NumPrefetches());
    }

    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        goto finalreset;
    }

    // By definition, the very last path is the terminator, and thus cannot have
    // an immediate predecessor.  Thus, there will not be a match of the previously
    // cached path.  We need to process the cached content and whatever we have
    // built to date as the current
    FindOrCreatePath(&s_prevPathCache, &cachedTemporalID, &cachedTemporalCount);
    FindOrCreatePath(s_currentPath, &finalTemporalID, &finalTemporalCount);

    if ((PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        g_PathProfiler->PassToStrataServer(cachedTemporalID, cachedTemporalCount, false);
        g_PathProfiler->PassToStrataServer(finalTemporalID, finalTemporalCount, true);
    }

finalreset: ;
    s_currentPath->ResetPath();
}

/// comparator for sorting of paths
static bool greaterHeat(const PathInfo_T& p1, const PathInfo_T& p2)
{
    return ((p1.tripCount * p1.pathHeat) > (p2.tripCount * p2.pathHeat));
}

/// Finish/shutdown IPath Profiling
/// Report out statistics, write the pathinfo db, clean up.
void DPathFinish(int retcode, void* dummy)
{
    PureDBRead  pathInfoDB;

    FlushLastPath();

    if (g_CmdOptions->DumpPathsToFile())
    {
        close(s_dumpFD);
    }

    /// Some experimental stuff (did my technique work?)
    /// Ans: sort-of.  Still need a dynamic_cast or static_cast
    DataPathProfiler* me = dynamic_cast<DataPathProfiler*>(g_PathProfiler->Self());
    me = dynamic_cast<DataPathProfiler*>(g_PathProfiler);
    me->SetSelf(me);


    // Reporting the results

    // Net statistics first
    // Should this be a csv format?  xml?
    std::string statsHdr;

    statsHdr = "Spreadsheet usable stats\n"
               "sshdr\tNumber\t\tNumber\tNumber\tMax\n"
               "sshdr\tAccesses\tPaths\tUnique\tPath\n"
               "sshdr\t\t\t\tPaths\tlength\n";
    g_Logger->LogIt(statsHdr);

    std::string statsValues;

    statsValues = "ssdata\t"
                  + ToString(s_DStats->TotalAccesses()) + "\t"     // number insns
                  + ToString(s_DStats->TotalCycles()) + "\t"    // number paths
                  + ToString(s_DStats->UniqueCycles()) + "\t"   // number unique paths
                  + ToString(s_DStats->MaxCycleLength()) + "\t" // max path path
                  + "\n";
    g_Logger->LogIt(statsValues);

    std::string histogramHdr;
    std::string histogramOutput;
    // Unique Path length distribution
    LengthHistogram_T   uniqueHist = s_DStats->UniqueLengthHistogram();

    histogramHdr = "\nUnique Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = uniqueHist.begin(); c_iter != uniqueHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);

    // Dynamic path length histogram
    LengthHistogram_T   dynHist = s_DStats->DynamicLengthHistogram();

    histogramHdr = "\nDynamic Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = dynHist.begin(); c_iter != dynHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);


    // DataCycle specific content, including access width distribution
    std::string     theData;
    theData += "\nTotalReads:\t" + ToString(s_DStats->TotalReads());
    theData += "\nTotalWrites:\t" + ToString(s_DStats->TotalWrites());
    theData += "\nTotalAtomics:\t" + ToString(s_DStats->TotalAtomics());
    theData += "\nTotalPrefetches:\t" + ToString(s_DStats->TotalPrefetches());
    theData += "\nTotalSpanPageReads:\t" + ToString(s_DStats->TotalSpanPageReads());
    theData += "\nTotalSpanPageWrites:\t" + ToString(s_DStats->TotalSpanPageWrites());
    theData += "\nMaximums per path:";
    theData += "\nMaxReads:\t" + ToString(s_DStats->MaxReads());
    theData += "\nMaxWrites:\t" + ToString(s_DStats->MaxWrites());
    theData += "\nMaxAtomics:\t" + ToString(s_DStats->MaxAtomics());
    theData += "\nMaxPrefetches:\t" + ToString(s_DStats->MaxPrefetches());
    theData += "\nMaxSpanPageReads:\t" + ToString(s_DStats->MaxSpanPageReads());
    theData += "\nMaxSpanPageWrites:\t" + ToString(s_DStats->MaxSpanPageWrites());
    theData += "\nMaxAccessWidth:\t" + ToString(s_DStats->MaxAccessWidth());
    g_Logger->LogIt(theData);

    std::string widthHistHdr;

    widthHistHdr = "\nAccess width distribution\n\tWidth\tCount\n";
    g_Logger->LogIt(widthHistHdr);

    std::string widthInfo;
    for (LengthHistogram_T::const_iterator c_iter = s_DStats->AccessHistogram().begin(); c_iter != s_DStats->AccessHistogram().end(); c_iter++)
    {
        widthInfo = widthInfo + "\t" + ToString(c_iter->first)
                              + "\t" + ToString(c_iter->second)
                              + "\n";
    }
    widthInfo = widthInfo + "\n";
    g_Logger->LogIt(widthInfo);


    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        return;
    }

    // write the pathinfo DB: statistics, info about the path vectors
    WritePathInfoDB();

    // More to report in path profiling and temporal profiling modes
    std::string cacheStats;

    cacheStats = "Highest immediate trip count: "
                + ToString(s_DStats->MaxImmediateTripCount())
                + "\tNumber of Cache hits: "
                + ToString(s_DStats->CachedHits())
                + "\n";
    g_Logger->LogIt(cacheStats);

    // And now some "interesting" stuff - for an IPath profiler, this would be the hot paths.
    // To do this: sort the pathmap by trip count after copying it into a sortable form.
    std::vector<PathInfo_T> pathsByCount;

    for (DataCycleToInfo_T::const_iterator c_iter = s_dataCycleToInfo->begin(); c_iter != s_dataCycleToInfo->end(); c_iter++)
    {
        pathsByCount.push_back(c_iter->second);
    }
    sort(pathsByCount.begin(), pathsByCount.end(), greaterHeat);

    // Now report the top 10 hottest paths - should this be tuned?
    // Again, boost::bimap may be less overhead to manipulate
    // Unfortunately, we either need to carry an entire map of the pathID --> the Vector
    // (lots and lots of memory in some cases) or we can read the values out of the
    // pathinfo DB.  Not as fast, but we aren't trying to grab all of them
    pathInfoDB.Open(g_PathProfiler->ProfilerPathInfoDBName());

    const std::string   dashSep("--------------------");
    std::string hotHdr;
    hotHdr = dashSep + dashSep + dashSep;
    hotHdr = hotHdr + "\nHottest 10 paths sorted by path heat (tripcount * number instructions)\n";
    g_Logger->LogIt(hotHdr);
    g_Logger->LogIt("\n\tPath Cycle contents (Access Mode: R = Read, W = Write, A = Atomic, P = prefetch)\n");
    g_Logger->LogIt("Address\tByte Width\tAccessMode\n");

    for (size_t i = 0; i < 10; i++)
    {
        PathInfo_T  info = pathsByCount[i];
        std::string marker;
        std::string     pathVectorKey;
        off_t           vecPos;
        size_t          vecLen;
        bool            foundVec;
        DataCycle_T     AddrVec;

        marker = dashSep + " Path # " + ToString(i + 1) + " " + dashSep + "\n";
        g_Logger->LogIt(marker);

        std::string summary;
        summary = "CycleID: " + ToString(info.cycleID)
                    + "\tTrips: " + ToString(info.tripCount)
                    + "\tHeat: " + ToString(info.pathHeat)
                    + "\tLength (accesess): " + ToString(info.pathLen)
                    + "\tDynamic instruction count: " + ToString(info.tripCount * info.pathLen)
                    + "\nDynamic path heat: " + ToString(info.tripCount * info.pathHeat)
                    + "\n\n";
        g_Logger->LogIt(summary);


        AddrVec.clear();
        pathVectorKey = s_dbDPathIDKey + ToString(info.cycleID);
        foundVec = pathInfoDB.Find(pathVectorKey, &vecPos, &vecLen);
        if (foundVec)
        {
            void*           dbRead;
            DataPathElt_T*  vecInfo;

            dbRead = pathInfoDB.Read(vecPos, vecLen);
            // dbRead is a buffer of bytes; we need to access it as if it were a vector
            // of addresses.  While it is a bit inefficient to build the vector from the
            // buffer, doing so is a bit less prone to errors than to simply doing a
            // C-style access knowing the datatype.

            // sanity check - we know how many elements there should be
            if ((info.pathLen * sizeof(DataPathElt_T)) != vecLen)
            {
                std::string errMsg;

                errMsg = "*** Error *** inconsistent information in PathInfo DB\n\tKey["
                         + pathVectorKey
                         + "] - expected "
                         + ToString(info.pathLen)
                         + " Elements, actual is: "
                         + ToString(vecLen / sizeof(DataPathElt_T))
                         + "\n";

                g_Logger->LogIt(errMsg);
            }

            vecInfo = reinterpret_cast<DataPathElt_T*>(dbRead);
            for (size_t j = 0; j < info.pathLen; j++)
            {
                std::string     vecContent;
                std::string     theMode("ERROR");

                    // Address  width    mode
                vecContent = "\t"
                             + ToString(vecInfo[j].address)
                             + "\t"
                             + ToString(vecInfo[j].width)
                             + "\t";

                // Convert the bitset into characters.  At instrumentation time
                // there is a separate call to extend the cycle when an instruction
                // has a write as well as a read (or two reads).
                // 
                if (s_accessBitsRead == (vecInfo[j].mode & s_accessBitsRead))
                {
                    vecContent += "R";
                }
                if (s_accessBitsWrite == (vecInfo[j].mode & s_accessBitsWrite))
                {
                    vecContent += "W";
                }
                if (s_accessBitsAtomicUpdate == (vecInfo[j].mode & s_accessBitsAtomicUpdate))
                {
                    vecContent += "A";
                }
                if (s_accessBitsPrefetch == (vecInfo[j].mode & s_accessBitsPrefetch))
                {
                    vecContent += "P";
                }
                vecContent += "\n";
                g_Logger->LogIt(vecContent);
            }   // Each address in the path

            pathInfoDB.ReadFree(dbRead);
        }
    g_Logger->LogIt("\n");
    }   // hottest i paths
}
