//
//  Copyright 2012-2013, Richard Gorton      All Rights Reserved
//
/// \file
/// Implementation of Istream Path Profiler (Instruction)
/// This file contains the execution-time content for the IPath profiler
///
/// \author
/// Richard Gorton
///

#include <limits>
#include <stdint.h>
#include <map>
#include <algorithm>
#include <cstring>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <zlib.h>

#include "util.h"
#include "pprof.h"
#include "CmdOptions.h"
#include "PathProfiler.h"
#include "Cycles.h"
#include "pstats.h"
#include "Logger.h"
#include "dbio.h"

/// C interfaces to the execution-time data collection for IPath profiling
extern "C" {
    void IPathInit(void*);
    void ConstructIPaths(void *BbAddr, uint32_t numBbInsns);
    void IPathFinish(int, void*);
}

extern PureDBWrite* g_BbDisasm_DB;

static InsnPath         s_prevPathCache;
static InsnPath*        s_currentPath = NULL;
static IPathStats*      s_IStats = NULL;

// the next one available to be assigned (these are unique)
static CycleID_T        s_nextCycleID = s_nullCycleID;

// s_rawIdx is the offset into the shared mem data area
static ShmArea_T    s_localSharedMem;
static size_t       s_rawIdx = 0;

// Only meaningful if we dump the pathid/tripcounts to a file
static int          s_dumpFD = -1;

/// The is the primary storage mechanism - a map: Path --> Info
static IPathtoInfo_T*     s_iPathToInfo = NULL;

/// Initialization of execution-time items
void IPathInit(void*)
{
    // This function is set up via a PIN_AddApplicationStartFunction()
    // callback which gets called once every image load.  Right now, this is now
    // how the istream gets instrumented, but this may change.

    static bool InitDone = false;

    if (InitDone)
    {
        return;
    }

    if (NULL == s_currentPath)
    {
        s_currentPath = new InsnPath();
    }
    if (NULL == s_currentPath)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_currentPath\n");
    }

    if (NULL == s_IStats)
    {
        s_IStats = new IPathStats();
    }
    if (NULL == s_IStats)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_IStats\n");
    }

    if ((PROFILEMODE_SAVEPATHS == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        // Construct the map to store the paths
        // IPath --> ID
        if (NULL == s_iPathToInfo)
        {
            s_iPathToInfo = new IPathtoInfo_T();
        }
        if (NULL == s_iPathToInfo)
        {
            LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_iPathToInfo\n");
        }
    }

    if (s_nullCycleID == s_nextCycleID)
    {
        s_nextCycleID = s_minCycleID;
    }

    // Initialization of the local shared memory copy.
    s_localSharedMem.dataHdr.version            = s_shm_verno;
    s_localSharedMem.dataHdr.operation          = OP_UNINITIALIZED;
    s_localSharedMem.dataHdr.collectionSet      = g_PathProfiler->CollectionSet() | g_PathProfiler->ProfileMode();
    s_localSharedMem.dataHdr.shm_flags          = SHMFLAG_UNINITALIZED;
    s_localSharedMem.dataHdr.payloadSize        = 0;
    s_localSharedMem.dataHdr.startInsnEpoch     = 0;
    s_localSharedMem.dataHdr.startPathEpoch     = 0;
    s_localSharedMem.dataHdr.startStrataEpoch   = 0;
    strcpy (s_localSharedMem.dataHdr.dataFileBaseName, g_PathProfiler->ProfilerBaseName().c_str());

    if (g_CmdOptions->DumpPathsToFile())
    {
        // Instead of writing to shared memory, we will end up writing the shared
        // memory buffers to a file
        int     oFlags = O_RDWR|O_CREAT|O_EXCL|O_TRUNC;
        mode_t  permissions = 0666;
        unlink(g_PathProfiler->StrataBuilderMapFile().c_str());
        s_dumpFD = open(g_PathProfiler->StrataBuilderMapFile().c_str(), oFlags, permissions);
        if (-1 == s_dumpFD)
        {
            LogAndAbort(__FUNCTION__, "Unable to create the file to dump the pathid/tripcounts into\n");
        }
    }

    InitDone = true;
}

/// Write the pathinfo db content out
static void WritePathInfoDB()
{
    CycleID_T   sanityCheck;
    PureDBWrite pathInfoDB;

    pathInfoDB.Open(g_PathProfiler->ProfilerPathInfoDBName());

    // Header
    pathInfoDB.Write(s_dbVerNumKey, s_dbVerNumValue);

    // Statistics.  See dbio.h for the conventions used
    // feels like some sort of table would be more appropriate
    pathInfoDB.Write("PathDictionary:1", s_dbStatsTotalCycles);
    pathInfoDB.Write(s_dbStatsTotalCycles, s_IStats->TotalCycles());

    pathInfoDB.Write("PathDictionary:2", s_dbStatsUniqueCycles);
    pathInfoDB.Write(s_dbStatsUniqueCycles, s_IStats->UniqueCycles());

    pathInfoDB.Write("PathDictionary:3", s_dbStatsMaxCycleLength);
    pathInfoDB.Write(s_dbStatsMaxCycleLength, s_IStats->MaxCycleLength());

    pathInfoDB.Write("PathDictionary:4", s_dbStatsMaxPathHeat);
    pathInfoDB.Write(s_dbStatsMaxPathHeat, s_IStats->MaxPathHeat());

    pathInfoDB.Write("PathDictionary:5", s_dbStatsMaxTripCount);
    pathInfoDB.Write(s_dbStatsMaxTripCount, s_IStats->MaxTripCount());

    pathInfoDB.Write("PathDictionary:6", s_dbStatsMaxImmediateTripCount);
    pathInfoDB.Write(s_dbStatsMaxImmediateTripCount, s_IStats->MaxImmediateTripCount());

    pathInfoDB.Write("PathDictionary:7", s_dbStatsCachedHits);
    pathInfoDB.Write(s_dbStatsCachedHits, s_IStats->CachedHits());

    pathInfoDB.Write("PathDictionary:8", s_dbStatsTotalInsns);
    pathInfoDB.Write(s_dbStatsTotalInsns, s_IStats->TotalInsns());

    pathInfoDB.Write("PathDictionary:9", s_dbStatsMaxPathInsnLen);
    pathInfoDB.Write(s_dbStatsMaxPathInsnLen, s_IStats->MaxPathInsnLen());

    pathInfoDB.Write("PathDictionary:10", s_dbStatsTotalBbs);
    pathInfoDB.Write(s_dbStatsTotalBbs, s_IStats->TotalBbs());

    uint64_t    numEntries = 10;
    pathInfoDB.Write("PathDictionary", numEntries);

    // And finally the contents of each unique path.
    // s_nullCycleID because we increment at the end of the loop
    sanityCheck = s_nullCycleID;

    for (IPathtoInfo_T::const_iterator c_iter = s_iPathToInfo->begin(); c_iter != s_iPathToInfo->end(); c_iter++)
    {
        std::string     theIDKey;
        std::string     theVectorKey;
        PathInfo_T      thisInfo;

        thisInfo = c_iter->second;
        theIDKey = s_dbPathInfoKey + ToString(thisInfo.cycleID);
        pathInfoDB.Write(theIDKey, thisInfo);

        InsnPath_T  BbVec = c_iter->first;
        theVectorKey = s_dbIPathIDKey + ToString(thisInfo.cycleID);

        /// Note: 6.2.3 of "The C++ Standard Library" by Josuttis notes that elements of a vector
        /// can be expected to be in contiguous memory.  Thus, we merely have to set the call
        /// to Write() up carefully, and do not need to perform a new/copy/delete.
        pathInfoDB.Write(const_cast<char*>(theVectorKey.c_str()),
                         reinterpret_cast<char*>(&BbVec[0]), 
                         BbVec.size() * sizeof(BbVec[0]));
        sanityCheck++;
    }

    if (s_IStats->UniqueCycles() != sanityCheck)
    {
        LogAndAbort(__FUNCTION__, "Consistency error: mismatch on number of unique paths vs. number in the map\n");
    }

    pathInfoDB.Close();
}

/// comparator for sorting of paths
static bool greaterHeat(const PathInfo_T& p1, const PathInfo_T& p2)
{
    return ((p1.tripCount * p1.pathHeat) > (p2.tripCount * p2.pathHeat));
}

/// A utility function to do the legwork of looking for a path in the maps, or creating
/// a new one.  Most of the time this is going to be passed the cached version, but for
/// the final flush, we also need to process the currently-being-built path.
/// \param[in]  the InstructionPath to be looked for
/// \param[out] the cycleID (a new one if it does not already exist)
/// \param[out] the paths TripCount
void FindOrCreatePath(InsnPath* findMe, CycleID_T* temporalID, TripCount_T* temporalCount)
{
    CycleID_T                   matchedID = 0;
    TripCount_T                 numTrips;
    IPathtoInfo_T::iterator     iter;

    numTrips = findMe->TripCount();

    iter = s_iPathToInfo->find(findMe->RawPathVector());
    if (iter != s_iPathToInfo->end())
    {
        // path pre-exists.
        matchedID = iter->second.cycleID;

        iter->second.tripCount += numTrips;

        if (s_IStats->MaxTripCount() < iter->second.tripCount)
        {
            s_IStats->SetMaxTripCount(iter->second.tripCount);
        }
    }   // was in IPathtoInfo map
    else
    {
        // New path.  Create the info and both of the addressvector/ID maps
        PathInfo_T  pInfo;

        matchedID = s_nextCycleID;

        pInfo.cycleID = s_nextCycleID;
        pInfo.tripCount = numTrips;
        pInfo.pathHeat = findMe->PathInsnLen();
        pInfo.pathLen = findMe->size();

        (*s_iPathToInfo)[findMe->RawPathVector()] = pInfo;

        // Need to allocate a new Cycleid.
        s_nextCycleID++;
        s_IStats->IncUniqueCycles(1);
        s_IStats->UniqueLengthHistogram()[findMe->size()]++;
    }
    *temporalID = matchedID;
    *temporalCount = numTrips;
}

/// Construct the paths and process them - this performs the bulk of the execution-time work
void ConstructIPaths(Address_T BbAddr, uint32_t numBbInsns)
{
    bool        quickMatch = false;
    CycleID_T   temporalID = s_nullCycleID;
    TripCount_T temporalCount = 0;

    s_IStats->IncTotalInsns(numBbInsns);
    s_IStats->IncTotalBbs(1);
    if (!s_currentPath->ExactMatch(BbAddr))
    {
        // add to the end of the potential cycle
        s_currentPath->ExtendIPath(BbAddr, numBbInsns);
        return;
    }

    // We found a match, it means we had a cycle; the level of processing
    // depends a lot on what the invocation settings were.

    // Always collect length information/statistics
    // Parent class
    s_IStats->IncTotalCycles(1);
    if (s_IStats->MaxCycleLength() < s_currentPath->size())
    {
        s_IStats->SetMaxCycleLength(s_currentPath->size());
    }
    s_IStats->DynamicLengthHistogram()[s_currentPath->size()]++;

    // For the simplistic approach, the Heat of a path is the length in instructions
    // A more interesting approach would be to calculate the number of cycles required
    // as computed on a static basis (that is, ignoring things like memory conflicts/stalls)
    PathHeat_T currPathHeat = s_currentPath->PathInsnLen();
    if (s_IStats->MaxPathHeat() > currPathHeat)
    {
        s_IStats->SetMaxPathHeat(currPathHeat);
    }

    // InsnPath statistics
    if (s_IStats->MaxPathInsnLen() < s_currentPath->PathInsnLen())
    {
        s_IStats->SetMaxPathInsnLen(s_currentPath->PathInsnLen());
    }

    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        goto resetfornext;
    }


    // For PROFILEMODE_SAVEPATHS, PROFILEMODE_STRATALENGTH, and PROFILEMODE_SAVESTRATA,
    // we need to perform further processing

    // Performance kicker:
    //  - while building the current path, keep a simple checksum as a quick check
    //    to possibly avoid the overhead of the find in the map; keep a copy of 
    //    the immediately previous vector/ID, so that if the checksum does match,
    //    we can try the immediate compare, and do a simple handoff.  If the
    //    checksum does not match, then yes we do have to do the find.  But tight
    //    loops are going to hit a lot.  Perhaps even a larger quickcache to
    //    catch a couple levels of nesting and/or simple control flow inside of
    //    loops in PROFILEMODE_SAVEPATHS (but not either of the STRATA collectors).

    if (s_prevPathCache.HashVal() == s_currentPath->HashVal())
    {
        if (s_prevPathCache.RawPathVector() == s_currentPath->RawPathVector())
        {
            quickMatch = true;
            s_IStats->IncCachedHits(1);
            s_prevPathCache.IncTripCount(1);
            if (s_IStats->MaxImmediateTripCount() < s_prevPathCache.TripCount())
            {
                s_IStats->SetMaxImmediateTripCount(s_prevPathCache.TripCount());
            }
        }
    }

    // Once the cache is being used, there are a number of cases to consider
    //  Pxn - Path <x>, trip #n, so PA1, PB1, PB2, PC1, etc.
    // Initial entry at program start
    //      PA1, PA2, PB1
    // Subsequent:
    //      PC1, PC2, PD1, PE1
    // Algorithm: if not in cache, find cache in the map
    //      if there, add in cache # trips
    //      if not, create with cache # trips

    // be sure not to create an entry for very first path
    if (!quickMatch && (0 < s_prevPathCache.size()))
    {
        FindOrCreatePath(&s_prevPathCache, &temporalID, &temporalCount);
    }

    if (!quickMatch)
    {
        s_prevPathCache = *s_currentPath;
        s_prevPathCache.SetTripCount(1);
    }

    // The temporal approach requires that we save a vector of CycleIDs.  The original
    // implementation took advantage of the fact that many paths are immediately repeated
    if ((PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        if (s_nullCycleID != temporalID)
        {
            // handoff through shared memory
            g_PathProfiler->PassToStrataServer(temporalID, temporalCount, false);
        }
    }

resetfornext: ;
    // Now that we have processed everything, reset the current path,
    // and start construction of a new path
    s_currentPath->ResetPath();
    s_currentPath->ExtendIPath(BbAddr, numBbInsns);

}


/// Pass a CycleID/tripCount to the Strata server process.
/// This happens via shared memory; we need to try to keep that memory as free from
/// contention on both sizes where possible.  Without being sure where potential bottlenecks
/// are, it is a reasonable first guess (due to the injection of the instrumentation code)
/// that the overhead incurred by the producer is larger than that of the consumer.
/// The original version kept a local copy of the buffer and built on that; when the
/// local copy got full, a memcpy into the shared memory area was done.  It might be
/// the case that a different set of semantics is better.  Analysis will be required
/// to determine that.  (aka To Be Determined later).
void InsnPathProfiler::PassToStrataServer (CycleID_T id, TripCount_T trips, bool finalFlush)
{
    PublicStratumElt_T  elt;

    elt.cycleID = id;
    elt.tripCount = trips;

    // buffer is full, or we need to complete the processing and shut down
    if ((s_rawIdx + sizeof(elt) > s_ShmemBufSize) || finalFlush)
    {
        char    opFlags;

        opFlags = OP_UNINITIALIZED;
        if (0 == s_IStats->SharedMemBufferCount())
        {
            opFlags |= OP_OPEN_APPEND;
        }
        if (finalFlush) {
            opFlags |= OP_FLUSH_CLOSE;
        }

        s_IStats->IncShmBufferCount(1);

        // using g_PathProfiler->m_Stats fails, but this->m_Stats.IncTotalBbs(1)
        // does work (presumably because "this" is in the correct context)

        // Build up a message for the server to analyze/compress
        s_localSharedMem.dataHdr.version        = s_shm_verno;
        s_localSharedMem.dataHdr.operation      = opFlags;
        s_localSharedMem.dataHdr.collectionSet  = g_PathProfiler->CollectionSet() | g_PathProfiler->ProfileMode();
        s_localSharedMem.dataHdr.payloadSize    = s_rawIdx;
        s_localSharedMem.dataHdr.startInsnEpoch = s_IStats->TotalInsns();
        s_localSharedMem.dataHdr.startPathEpoch = s_IStats->TotalCycles();
        s_localSharedMem.dataHdr.startStrataEpoch = 0;

        if (g_CmdOptions->DumpPathsToFile())
        {
            // Simply mark this buffer as "filled" and write the entire area to file
            // So our physical file will be a multiple of ShmArea_T bytes
            s_localSharedMem.dataHdr.shm_flags = SHMFLAG_FILLED;
            write(s_dumpFD, &s_localSharedMem, sizeof(ShmArea_T));
        }
        else
        {
            // Send the message to the server, once server is available
            // Technique for debugging the strata server:
            //  Break on the MarkDataFilled statement in the path generator;
            //  In the strata server, set a break on the WaitUntilReadyToFill
            WaitUntilReadyToFill (g_PathProfiler->StrataSharedMem());
            memcpy ((void*) g_PathProfiler->StrataSharedMem(), (void*) &s_localSharedMem, sizeof(s_localSharedMem));
            MarkDataFilled (const_cast<ShmArea_T*>(g_PathProfiler->StrataSharedMem()));
        }

        // reset for further use
        s_rawIdx = 0;
    }

        // copy new info into buffer
    memcpy ((void*) &s_localSharedMem.rawData[s_rawIdx], (void*) &elt, sizeof(elt));
    s_rawIdx += sizeof(elt);

}


/// Flush the last path (and flush the strata as well if relevant)
/// In this case, we will need to flush the cached content, along with the
/// "current" content.
static void FlushLastPath()
{
    CycleID_T   cachedTemporalID = s_nullCycleID;
    CycleID_T   finalTemporalID = s_nullCycleID;
    TripCount_T cachedTemporalCount = 0;
    TripCount_T finalTemporalCount = 0;

    // Whatever partial path is there is now "complete"
    s_IStats->IncTotalCycles(1);
    if (s_IStats->MaxCycleLength() < s_currentPath->size())
    {
        s_IStats->SetMaxCycleLength(s_currentPath->size());
    }
    s_IStats->DynamicLengthHistogram()[s_currentPath->size()]++;

    // For the simplistic approach, the Heat of a path is the length in instructions
    // A more interesting approach would be to calculate the number of cycles required
    // as computed on a static basis (that is, ignoring things like memory conflicts/stalls)
    PathHeat_T currPathHeat = s_currentPath->PathInsnLen();
    if (s_IStats->MaxPathHeat() > currPathHeat)
    {
        s_IStats->SetMaxPathHeat(currPathHeat);
    }

    // InsnPath statistics
    if (s_IStats->MaxPathInsnLen() < s_currentPath->PathInsnLen())
    {
        s_IStats->SetMaxPathInsnLen(s_currentPath->PathInsnLen());
    }

    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        goto finalreset;
    }

    // By definition, the very last path is the terminator, and thus cannot have
    // an immediate predecessor.  Thus, there will not be a match of the previously
    // cached path.  We need to process the cached content and whatever we have
    // built to date as the current
    FindOrCreatePath(&s_prevPathCache, &cachedTemporalID, &cachedTemporalCount);
    FindOrCreatePath(s_currentPath, &finalTemporalID, &finalTemporalCount);

    if ((PROFILEMODE_STRATALENGTH == g_PathProfiler->ProfileMode())
        || (PROFILEMODE_SAVESTRATA == g_PathProfiler->ProfileMode()))
    {
        g_PathProfiler->PassToStrataServer(cachedTemporalID, cachedTemporalCount, false);
        g_PathProfiler->PassToStrataServer(finalTemporalID, finalTemporalCount, true);
    }

finalreset: ;
    s_currentPath->ResetPath();
}

/// Finish/shutdown IPath Profiling
/// Report out statistics, write the pathinfo db, clean up.
void IPathFinish(int retcode, void* dummy)
{
    PureDBRead  disAsmDB;
    PureDBRead  pathInfoDB;

    FlushLastPath();

    if (g_CmdOptions->DumpPathsToFile())
    {
        close(s_dumpFD);
    }

    // Close out writing of the disasm db
    g_BbDisasm_DB->Close();
    g_BbDisasm_DB = NULL;

    /// Some experimental stuff (did my technique work?)
    /// Ans: sort-of.  Still need a dynamic_cast or static_cast
    InsnPathProfiler* me = dynamic_cast<InsnPathProfiler*>(g_PathProfiler->Self());
    me = dynamic_cast<InsnPathProfiler*>(g_PathProfiler);
    me->SetSelf(me);


    // Reporting the results

    // Net statistics first
    // Should this be a csv format?  xml?
    std::string statsHdr;

    statsHdr = "Spreadsheet usable stats\n"
               "sshdr\tNumber\tNumber\tNumber\tNumber\tMax\tMax\tMean\n"
               "sshdr\tInsns\tBbs\tPaths\tUnique\tInsns\tBbs\tBb\n"
               "sshdr\t\t\t\tPaths\tper\tper\tPath\n"
               "sshdr\t\t\t\t\tPath\tPath\tLen\n";
    g_Logger->LogIt(statsHdr);

    // It is not possible to have zero paths...
    PathLen_T   meanPathBbLen = s_IStats->TotalBbs() / s_IStats->TotalCycles();

    std::string statsValues;

    statsValues = "ssdata\t"
                  + ToString(s_IStats->TotalInsns()) + "\t"     // number insns
                  + ToString(s_IStats->TotalBbs()) + "\t"       // number bbs
                  + ToString(s_IStats->TotalCycles()) + "\t"    // number paths
                  + ToString(s_IStats->UniqueCycles()) + "\t"   // number unique paths
                  + ToString(s_IStats->MaxPathInsnLen()) + "\t" // max insn on a path
                  + ToString(s_IStats->MaxCycleLength()) + "\t" // max bbs on a path
                  + ToString(meanPathBbLen)                     // mean bb path len (dynamic)
                  + "\n";
    g_Logger->LogIt(statsValues);

    std::string histogramHdr;
    std::string histogramOutput;
    // Unique Path length distribution
    LengthHistogram_T   uniqueHist = s_IStats->UniqueLengthHistogram();

    histogramHdr = "\nUnique Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = uniqueHist.begin(); c_iter != uniqueHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);

    // Dynamic path length histogram
    LengthHistogram_T   dynHist = s_IStats->DynamicLengthHistogram();

    histogramHdr = "\nDynamic Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = dynHist.begin(); c_iter != dynHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);

    if (PROFILEMODE_PATHLENGTH == g_PathProfiler->ProfileMode())
    {
        return;
    }

    // write the pathinfo DB: statistics, info about the path vectors
    WritePathInfoDB();

    // More to report in path profiling and temporal profiling modes
    std::string cacheStats;

    cacheStats = "Highest immediate trip count: "
                + ToString(s_IStats->MaxImmediateTripCount())
                + "\tNumber of Cache hits: "
                + ToString(s_IStats->CachedHits())
                + "\n";
    g_Logger->LogIt(cacheStats);

    // And now some "interesting" stuff - for an IPath profiler, this would be the hot paths.
    // To do this: sort the pathmap by trip count after copying it into a sortable form.
    std::vector<PathInfo_T> pathsByCount;

    for (IPathtoInfo_T::const_iterator c_iter = s_iPathToInfo->begin(); c_iter != s_iPathToInfo->end(); c_iter++)
    {
        pathsByCount.push_back(c_iter->second);
    }
    sort(pathsByCount.begin(), pathsByCount.end(), greaterHeat);

    // Now report the top 10 hottest paths - should this be tuned?
    // Again, boost::bimap may be less overhead to manipulate
    // Unfortunately, we either need to carry an entire map of the pathID --> the Vector
    // (lots and lots of memory in some cases) or we can read the values out of the
    // pathinfo DB.  Not as fast, but we aren't trying to grab all of them
    pathInfoDB.Open(g_PathProfiler->ProfilerPathInfoDBName());
    disAsmDB.Open(g_PathProfiler->ProfilerDisAsmDBName());

    const std::string   dashSep("--------------------");
    std::string hotHdr;
    hotHdr = dashSep + dashSep + dashSep;
    hotHdr = hotHdr + "\nHottest 10 paths sorted by path heat (tripcount * number instructions)\n";
    g_Logger->LogIt(hotHdr);

    for (size_t i = 0; i < 10; i++)
    {
        PathInfo_T  info = pathsByCount[i];
        std::string marker;
        std::string     pathVectorKey;
        off_t           vecPos;
        size_t          vecLen;
        bool            foundVec;
        InsnPath_T      BbVec;

        marker = dashSep + " Path # " + ToString(i + 1) + " " + dashSep + "\n";
        g_Logger->LogIt(marker);

        std::string summary;
        summary = "CycleID: " + ToString(info.cycleID)
                    + "\tTrips: " + ToString(info.tripCount)
                    + "\tHeat: " + ToString(info.pathHeat)
                    + "\tLength (Bbs): " + ToString(info.pathLen)
                    + "\nDynamic instruction count: " + ToString(info.tripCount * info.pathHeat)
                    + "\tDynamic Bb count: " + ToString(info.tripCount * info.pathLen)
                    + "\n\n";
        g_Logger->LogIt(summary);

        g_Logger->LogIt("\n\tPath Disassembly:\n");

        BbVec.clear();
        pathVectorKey = s_dbIPathIDKey + ToString(info.cycleID);
        foundVec = pathInfoDB.Find(pathVectorKey, &vecPos, &vecLen);
        if (foundVec)
        {
            void*       dbRead;
            Address_T*  vecInfo;

            dbRead = pathInfoDB.Read(vecPos, vecLen);
            // dbRead is a buffer of bytes; we need to access it as if it were a vector
            // of addresses.  While it is a bit inefficient to build the vector from the
            // buffer, doing so is a bit less prone to errors than to simply doing a
            // C-style access knowing the datatype.

            // sanity check - we know how many elements there should be
            if ((info.pathLen * sizeof(Address_T)) != vecLen)
            {
                std::string errMsg;

                errMsg = "*** Error *** inconsistent information in PathInfo DB\n\tKey["
                         + pathVectorKey
                         + "] - expected "
                         + ToString(info.pathLen)
                         + " Elements, actual is: "
                         + ToString(vecLen / sizeof(Address_T))
                         + "\n";

                g_Logger->LogIt(errMsg);
            }

            vecInfo = reinterpret_cast<Address_T*>(dbRead);
            for (size_t j = 0; j < info.pathLen; j++)
            {
                BbVec.push_back(vecInfo[j]);
            }

            pathInfoDB.ReadFree(dbRead);
        }

        for (InsnPath_T::const_iterator c_iter = BbVec.begin(); c_iter != BbVec.end(); c_iter++)
        {
            char        KeyStr[128];
            char        uncompressBuf[1024*1024];
            size_t      uncompressLen = sizeof(uncompressBuf);
            off_t       findpos;
            size_t      findlen;
            bool        foundKey = false;

            // Build the key
            snprintf (KeyStr, sizeof(KeyStr), "%p", *c_iter);

            foundKey = disAsmDB.Find(KeyStr, &findpos, &findlen);
            if (foundKey)
            {
                void*   dbRead;

                dbRead = disAsmDB.Read(findpos, findlen);
                uncompress((Bytef*)uncompressBuf, &uncompressLen, (Bytef*)dbRead, findlen);
                disAsmDB.ReadFree(dbRead);
            }

            std::string decode(uncompressBuf, 0, uncompressLen);
            g_Logger->LogIt(decode);

        }   // each Bb
    }   // hottest i paths
}
