//
//  Copyright 2012-2013, Richard Gorton      All Rights Reserved
//
/// \file
/// Implementation of Istream Path Profiler (Instruction)
/// This file contains the instrumentation-time content for the IPath profiler
///
/// \author
/// Richard Gorton
///

#include <limits.h>
#include <stdint.h>
#include <zlib.h>

#include <limits>
#include <map>

#include "pin.H"
#include "instlib.H"
#include "xed-interface.h"

#include "util.h"
#include "pprof.h"
#include "CmdOptions.h"
#include "PathProfiler.h"
#include "Cycles.h"
#include "pstats.h"
#include "dbio.h"

/// Keep a separate DB which stores the disassembly.
/// This is a bit different than the prior approach wherein the disassembly
/// was kept in the same DB as the paths themselves
extern "C" {
    /// C interfaces to the execution-time data collection for IPath profiling
    void IPathInit(void*);
    void ConstructIPaths();
    void IPathFinish(int, void*);
}

PureDBWrite*    g_BbDisasm_DB = NULL;

////
/// create an entry (if necessary) in the db of the form:
/// Key = <bbaddr> - expressed with sprintf (buf, "%p", <bbaddr>)
/// Value = compressed disassembly (by PIN) of said Bb
///
/// MAJOR ASSUMPTION: the number of instructions for a BbAddr NEVER, EVER changes
/// but this may not be true, based upon PIN documentation.
///
static bool StoreDisAsm (const Address_T& BbAddr, std::string& disassembly)
{
    int         KeyLen;
    char        KeyStr[128];
    size_t      compDisLen;
    char        compressBuf[1024*1024];
    int         compres;
    bool        result = true;

        // Build the key
    snprintf (KeyStr, sizeof(KeyStr), "%p", BbAddr);
    KeyLen = strlen (KeyStr);

        // Compress the input: length + 1 for trailing null; 9 is max compression
    compDisLen = sizeof(compressBuf);
    compres = compress2 ((Bytef*)compressBuf,
                         &compDisLen,               // An output param
                         (Bytef*)disassembly.c_str(),
                         disassembly.size(),
                         9);

    result = g_BbDisasm_DB->Write(&KeyStr[0], &compressBuf[0], compDisLen);

    return result;
}

///
/// Possible improvements: Construct and store the disassembly in the Fini()
/// function for performance:
///      - Linux Filesystems try to keep the DB it in memory, which contributes to the
///      in-memory footprint of the application (gobmk is already a multi-Gb consumer)
///      - We don't spend cycles re-writing the disassembly multiple times, because we
///      do have all of the Bbs.
///      - Since we do this afterwards, we don't take the hit during execution.
/// This looks to be possible using the XED features in PIN.
///
void GenerateIPathProfile (TRACE trace, void* v)
{
    for (BBL bblist = TRACE_BblHead(trace); BBL_Valid (bblist); bblist = BBL_Next(bblist)) {
        string  BbDecode;
        char    Blob[128];
        ADDRINT BbAddr;
        unsigned int    BbInsns;
        std::string     rtnName;

        BbAddr = BBL_Address (bblist);
        BbInsns = BBL_NumIns (bblist);
        rtnName = RTN_FindNameByAddress (BbAddr);

        for (INS disIns = BBL_InsHead (bblist); INS_Valid (disIns); disIns = INS_Next (disIns)) {

            if (disIns == BBL_InsHead (bblist)) {

                snprintf (Blob, sizeof(Blob), "\t# 0x%lx\t%d Instructions",
                            static_cast<uint64_t>(BbAddr),
                            BbInsns);
                BbDecode = Blob;
                BbDecode = BbDecode + "\tRoutine: [" + rtnName + "]\n";
            }
            BbDecode += "\t\t" + INS_Disassemble(disIns) + "\n";
        }       // All INS in the Bb.
        BbDecode += "\n";

        //
        // MAJOR ASSUMPTION: the number of instructions for a BbAddr NEVER, EVER changes
        // but this may not be true, based upon PIN documentation.
        //

        if (!StoreDisAsm (reinterpret_cast<Address_T>(BbAddr), BbDecode))
        {
            cerr << "Fatal problem - unable to save disassembly - aborting\n"
                 << endl;
            exit(1);
        }
        BBL_InsertCall (bblist, IPOINT_ANYWHERE, ConstructIPaths,
                        IARG_ADDRINT, BbAddr,
                        IARG_UINT32, BbInsns,
                        IARG_END);
    }   // Each BBL
}

/// Initialize
bool InsnPathProfiler::Initialize()
{
    PIN_InitSymbols();

    g_BbDisasm_DB = new PureDBWrite();
    g_BbDisasm_DB->Open(g_PathProfiler->ProfilerDisAsmDBName());

    return true;
}

/// InstrumentApp
bool InsnPathProfiler::InstrumentApp()
{
    PIN_AddApplicationStartFunction(IPathInit, 0);

      // Count Bbs, Add path profiling
    TRACE_AddInstrumentFunction(GenerateIPathProfile, 0);

    PIN_AddFiniFunction(IPathFinish, 0);

    return true;
}

bool InsnPathProfiler::StartProfiling()
{

    // Once called, control never returns.  Because of this StopProfiling and Finish
    // we can't perform any operations of consequence in ::StopProfiling and ::Finish
    // unless we call them as PIN_AddFiniFunction() items.
    PIN_StartProgram();

    return true;
}

bool InsnPathProfiler::StopProfiling()
{
    return true;
}

bool InsnPathProfiler::Finish()
{
    return true;
}

