//
// Copyright, 2007-2013 Richard Gorton      All Rights Reserved
//
// stratum server to compress cycles of paths.  Run as a separate process to enable
// asynchronous offload of computes to a separate core.  The incoming data is comprised
// of path ids.  Much like a hot path profilers, this server will analyze the incoming
// content and write the algorithmically reduced content to a compression server.
// We define a stratum to be a cycle of [repeated] paths.
// If we have a path (a dynamic cycle of addresses) A with an immediate repeat count of 100,
// be expressed as A(100).  That is: the cycle of Bbs embodied by A was contiguously executed
// exactly 100 times.
// A stratum is then a cycle of repeated paths, including trip count.  For example:
// Paths: --> Stratum contents
//  A(100), B(12), C(17), A(99), B(12) --> A(100), B(12), C(17), A(99)
//  A(100), B(12), C(17), A(100), B(12) --> A(100), B(12), C(17)
//
// We also extend the concept beyond that of the addresses of Bbs, and rather apply
// it more generally to mere memory addresses.  This enables us to capture (in a hopefully
// very powerful and compact form) data address traces.
//
// Assumptions:
//      Only one sender of data at a time
//      Uses sysV shmem as the data transfer method
//      Started after the compression server by the PIN-based analysis tool
//
// Some statistics about the strata from the original experiment work
//      Number of unique strata:         2862441 ref/bzip2
//      Longest strata (in paths):           444 test/wrf train/wrf ref/wrf
//      Highest continuous tripcount: 5121578643 ref/sphinx
//

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <zlib.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include <ostream>
#include <string>

#include "pprof.h"
#include "util.h"
#include "shmio.h"
#include "dbio.h"
#include "Logger.h"
#include "Cycles.h"
#include "pstats.h"

ProfileLogger*      g_Logger = NULL;
char*               g_argV0;

static struct DF_INFO   s_dataPair;
char                g_strataMapFileName[PATH_MAX];
char                g_compressMapFileName[PATH_MAX];

// Default to not operating on a dumpfile as input
bool                g_dumpMode = false;
int                 g_dumpFD = -1;

// Default to collecting and saving everything, not just collecting length
// histogram information.  This is both faster, and has a much smaller memory
// footprint, depending upon the application.
bool                g_saveStrata = true;

// the next one available to be assigned (these are unique)
static SID_T        s_nextSID = static_cast<SID_T>(s_nullCycleID);


// Shared memory regions for passing processed content into here, and subsequently
// to the compression process
static ShmArea_T*   s_collectorSharedMem = NULL;
static ShmArea_T*   s_compressSharedMem = NULL;

// working copies of the incoming/outgoing shared mem.  To avoid stalling other processes,
// we make a working copy of the entire shared memory area and work with that,
// rather than the "live" area.
static ShmArea_T    s_incomingShM;
static ShmArea_T    s_outboundShM;
static size_t       s_rawIdx = 0;

static std::string  s_cSetPrefix("UNINITIALIZED");

static StrataCycle  s_prevStrataCache;
static StrataCycle* s_currentStrataCycle = NULL;
static CycleStats*  s_strataStats = NULL;

static SCycleToSInfo_T*       s_StrataCycletoInfo = NULL;
static std::string  s_strataInfoDBName("UNINITIALIZED");
// For purposes of saving memory, we will not insert all cycles into the map
// We will merely mark them as new, and write them directly out
// bzip2 ref run @ 16 - killed at 15+Gb mem consumption
//               @ 8 - seems to use less than 1Gb
//               @ 12 - killed at 16+Gb memory
//  Interesting.  Must be some sort of combinatorial cliff
static const size_t s_neverMap = 8;

//
// Overview:
//  A stratum layer is a cycle of repeated paths.  We collect those, and store
//  the information in a pureDB database.  Read the incoming data from shared
//  memory, and construct the cycles.  Pass the raw final data to the compression
//  server (also via shared memory)  The final content is a series of Strata IDs with
//  trip counts.
//

/// Operator overloads to enable simple comparisons of the vectors based upon
/// an article on cplusplus.com: "compare vectors for equality"
/// I tried putting it inside the class definition for StrataCycle, but that fails
/// at compile time "error: ... must take exactly one argument"
/// there is a response that it needs to be overloaded at global scope.
/// putting it into a .h file resulted in errors about it being multiply defined.
/// Need an operator for "==" to enable:
///         if (searchForElt == *where)
bool operator == (const StratumElt_T& p1, const StratumElt_T& p2)
{
    return ((p1.Elt_ID == p2.Elt_ID)
            && (p1.Elt_Trips == p2.Elt_Trips));
}

/// this overload is needed for map insertion:
///         (*s_StrataCycletoInfo)[findMe->RawPathVector()] = <value>;
bool operator < (const StratumElt_T& p1, const StratumElt_T& p2)
{
    if (p1.Elt_ID < p2.Elt_ID)
    {
        return true;
    }
    if ((p1.Elt_ID == p2.Elt_ID)
            && (p1.Elt_Trips < p2.Elt_Trips))
    {
        return true;
    }
    return false;
}

bool StrataCycle::ExactMatch(const StratumElt_T searchForElt)
{
    StrataCycle_T::const_iterator  where;

    for (where = m_pathVec.begin(); where != m_pathVec.end(); where++)
    {
#if 0
        // Use this if there is no overload for the "==" operator
        if ((searchForElt.Elt_ID == where->Elt_ID)
            && (searchForElt.Elt_Trips == where->Elt_Trips))
#else
        if (searchForElt == *where)
#endif
        {
            return true;
        }
    }
    return false;
}


// \returns     A potential exit code so it is possible to do a "return usage()" from main().
int usage()
{
    std::cerr << "Invocation: "
              << g_argV0
              << " <stratum mmapfilename> <compressor mmapfilename>\n"
              << "OR: "
              << g_argV0
              << " <input dumpfile> (no output)\n";
    return 1;
}

// Handling a fatal situation - log to stderr and to the logfile (if possible)
// C style interface here, globally accessible
extern "C" void LogAndAbort(const char* whence, const char* simpleMessage)
{
    std::string fullMessage;

    fullMessage = whence;
    fullMessage += " ";
    fullMessage += simpleMessage;
    fullMessage += "\n";
    std::cerr << fullMessage;
    if (NULL != g_Logger)
    {
        g_Logger->LogIt(fullMessage);
    }
    usage();
    exit(1);
}

static void cleanup()
{
    if (!g_dumpMode)
    {
        unlink (g_strataMapFileName);
    }
    unlink (g_compressMapFileName);
}

static void cleanup_exit(int retcode)
{
    fprintf (stderr, "Aborting\n");
    cleanup();
    exit (retcode);
}

void StrataInit(const char* baseName)
{
    static bool initDone = false;
    std::string logFileName;
    std::string dbBaseName;
    std::string rawStrataFile;

    if (initDone)
    {
        LogAndAbort(__FUNCTION__, "Initialization already done - aborting\n");
    }

    dbBaseName = std::string(baseName) + "." + s_cSetPrefix + s_strataSuffix;
    logFileName = dbBaseName + s_logSuffix;
    s_strataInfoDBName = dbBaseName + s_infoSuffix + s_dbSuffix;
    g_Logger->SetLogFileName(logFileName);

    // the raw strata file is the content which we will be compressing
    rawStrataFile = dbBaseName + s_rawSuffix;
    strcpy (s_outboundShM.dataHdr.dataFileBaseName, rawStrataFile.c_str());

    if (NULL == s_currentStrataCycle)
    {
        s_currentStrataCycle = new StrataCycle();
    }
    if (NULL == s_currentStrataCycle)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_currentStrataCycle\n");
    }

    if (NULL == s_strataStats)
    {
        s_strataStats = new CycleStats();
    }
    if (NULL == s_strataStats)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_strataStats\n");
    }

    // maps for the strata cycle to/from id, the info for some statistics about
    // each SID (StrataCycle ID)
    if (NULL == s_StrataCycletoInfo)
    {
        s_StrataCycletoInfo = new SCycleToSInfo_T();
    }
    if (NULL == s_StrataCycletoInfo)
    {
        LogAndAbort(__FUNCTION__, "Unable to allocate memory for s_StrataCycletoInfo\n");
    }

    if (s_nullCycleID == s_nextSID)
    {
        s_nextSID = s_minCycleID;
    }

    initDone = true;
}

/// Final handoff - the ID/tripcount pair is passed to the compression server
void PassToCompressionServer(const SID_T& id, const STrip_T& trips, bool finalFlush)
{
    StratumElt_T    elt;

    elt.Elt_ID = id;
    elt.Elt_Trips = trips;

    // buffer is full, or we need to complete the processing and shut down
    if ((s_rawIdx + sizeof(elt) > s_ShmemBufSize) || finalFlush)
    {
        char    opFlags;
    
        opFlags = OP_UNINITIALIZED;
        if (0 == s_strataStats->SharedMemBufferCount())
        {
            opFlags |= OP_OPEN_APPEND;
        }
        if (finalFlush) {
            opFlags |= OP_FLUSH_CLOSE;
        }

        s_strataStats->IncShmBufferCount(1);

        // Build up a message for the server to analyze/compress
        s_outboundShM.dataHdr.version           = s_shm_verno;
        s_outboundShM.dataHdr.operation         = opFlags;
        s_outboundShM.dataHdr.collectionSet     = s_incomingShM.dataHdr.collectionSet;
        s_outboundShM.dataHdr.payloadSize       = s_rawIdx;
        s_outboundShM.dataHdr.startInsnEpoch    = s_incomingShM.dataHdr.startInsnEpoch;
        s_outboundShM.dataHdr.startPathEpoch    = s_incomingShM.dataHdr.startPathEpoch;
        s_outboundShM.dataHdr.startStrataEpoch  = s_strataStats->TotalCycles();

        if (!g_saveStrata)
        {
            // we'll copy the "content" in place, but do nothing with it.
            s_outboundShM.dataHdr.payloadSize = 0;
        }

        if (g_dumpMode)
        {
            // In read-the-dump-file mode, there is no handoff to the compressor
            // via shared memory.
        }
        else
        {
            // In non-dump mode, Send the message to the server, once available
            // Technique for debugging the strata server:
            //  Break on the MarkDataFilled statement in the path generator;
            //  In the strata server, set a break on the WaitUntilReadyToFill

            // Because the "shutdown" mechanism for the compressor happens via
            // a flag in the shared memory, we do have to pass the header even in
            // the case of not saving the strata content
            WaitUntilReadyToFill (s_compressSharedMem);
            memcpy ((void*) s_compressSharedMem, (void*) &s_outboundShM, sizeof(s_outboundShM));
            MarkDataFilled (const_cast<ShmArea_T*>(s_compressSharedMem));
        }

        // reset for further use
        s_rawIdx = 0;
    }

        // copy new info into buffer
    memcpy ((void*) &s_outboundShM.rawData[s_rawIdx], (void*) &elt, sizeof(elt));
    s_rawIdx += sizeof(elt);

}

/// A utility function to do the legwork of looking for a path in the maps, or creating
/// a new one.  Most of the time this is going to be passed the cached version, but for
/// the final flush, we also need to process the currently-being-built path.
/// For purposes of keeping memory consumption under control, we basically treat "long"
/// cycles as one-offs and do not try to store them in the map.  This behavior is specific
/// to the strata builder.
/// \param[in]  the StrataCycle to be looked for
/// \param[out] the cycleID (a new one if it does not already exist)
/// \param[out] the paths TripCount
void FindOrCreatePath(StrataCycle* findMe, SID_T* temporalID, STrip_T* temporalCount)
{

    SID_T       matchedID = 0;
    STrip_T     numTrips;
    bool            makeNew;
    bool            underSizeLimit;
    SCycleToSInfo_T::iterator iter;

    numTrips = findMe->TripCount();
    makeNew = false;
    underSizeLimit = false;

    if (!g_saveStrata)
    {
        // the map is empty, and should remain empty.  So we skip to the end
        goto done;
    }

    if (findMe->size() > s_neverMap)
    {
        makeNew = true;
        // In order to keep our statistics correct, track the number of
        // 'new' items never put into the map.
        s_strataStats->IncUnmappedLongCycles(1);
        s_strataStats->UniqueLengthHistogram()[findMe->size()]++;
    }
    else
    {
        // We are under the maximum size for mapping, so we can go
        // check for the cycle in the map
        underSizeLimit = true;

        iter = s_StrataCycletoInfo->find(findMe->RawPathVector());
        if (iter != s_StrataCycletoInfo->end())
        {
            // path pre-exists.
            matchedID = iter->second.StrID;

            iter->second.trips += numTrips;

            if (s_strataStats->MaxTripCount() < iter->second.trips)
            {
                s_strataStats->SetMaxTripCount(iter->second.trips);
            }
        }   // was in SCycleToSID
        else
        {
            makeNew = true;
        }
    }

    if (makeNew)
    {
        // New path.  Create the info and both of the addressvector/ID maps
        matchedID = s_nextSID;

        if (underSizeLimit)
        {
            StratumInfo_T   SInfo;

            SInfo.StrID = s_nextSID;
            SInfo.trips = numTrips;

            (*s_StrataCycletoInfo)[findMe->RawPathVector()] = SInfo;
        }

        // Need to allocate a new cycleid.
        s_nextSID++;
        s_strataStats->IncUniqueCycles(1);
        s_strataStats->UniqueLengthHistogram()[findMe->size()]++;
    }

done: ;
    *temporalID = matchedID;
    *temporalCount = numTrips;
}


// Extend the strata layer
void ConstructStrataLayer(const StratumElt_T& theElt)
{
    bool        quickMatch      = false;
    SID_T       temporalID      = s_nullCycleID;
    STrip_T     temporalCount   = 0;

    if (!s_currentStrataCycle->ExactMatch(theElt))
    {
        s_currentStrataCycle->ExtendCycle(theElt);
        return;
    }

    // A cycle completed

    // Collect statistics
    s_strataStats->IncTotalCycles(1);   // could hoist this out, but be roughly consistent with
                                        // ipath construction
    if (s_strataStats->MaxCycleLength() < s_currentStrataCycle->size())
    {
        s_strataStats->SetMaxCycleLength(s_currentStrataCycle->size());
    }
    s_strataStats->DynamicLengthHistogram()[s_currentStrataCycle->size()]++;

    // We are here because we are in PROFILEMODE_TEMPORAL, so do the caching
    // maneuver that was done for ipath collection

    if (s_prevStrataCache.HashVal() == s_currentStrataCycle->HashVal())
    {
        if (s_prevStrataCache.RawPathVector() == s_currentStrataCycle->RawPathVector())
        {
            quickMatch = true;
            s_strataStats->IncCachedHits(1);
            s_prevStrataCache.IncTripCount(1);
            if (s_strataStats->MaxImmediateTripCount() < s_prevStrataCache.TripCount())
            {
                s_strataStats->SetMaxImmediateTripCount(s_prevStrataCache.TripCount());
            }
        }   // vector match
    }   // hash match

    // almost cut-and-paste from ipath
    // There are some differences, in that we are not trying to further
    // compress into cycles of strata cycles.

    // be sure not to create an entry for very first path
    if (!quickMatch && (0 < s_prevStrataCache.size()))
    {
        FindOrCreatePath(&s_prevStrataCache, &temporalID, &temporalCount);
    }

    if (!quickMatch)
    {
        // Copy the values from s_currentStrataCycle into the cache.  Do this
        // "by hand" to avoid some of the allocates/releases incurred by
        //      s_prevStrataCache = *s_currentStrataCycle;
        s_prevStrataCache.SetCycleID(s_currentStrataCycle->CycleID());
        s_prevStrataCache.SetNumElements(s_currentStrataCycle->NumElements());
        s_prevStrataCache.SetTripCount(1);
        s_prevStrataCache.SetHashVal(s_currentStrataCycle->HashVal());
        s_prevStrataCache.SetRawPathVector(s_currentStrataCycle->RawPathVector());
    }

    // The temporal approach requires that we save a vector of cycleIDs.  The original
    // implementation took advantage of the fact that many paths are immediately repeated
    // We have taken this to the [current] limit of our technique, so it is time to
    // to the final handoff to the compression server
    if (s_nullCycleID != temporalID)
    {
        PassToCompressionServer(temporalID, temporalCount, false);
    }

    // Now that we have processed everything, reset the current path,
    // and start construction of a new path
    s_currentStrataCycle->ResetCycle();
    s_currentStrataCycle->ExtendCycle(theElt);

}

/*
 * Analyze and algorithmically compress the strata (cyclical path sequences)
 * Generate a strata info database file containing the details of each strata.
 */
void processPayload (ShmArea_T* memData)
{
    PublicStratumElt_T* incomingElts;
    uint64_t        i;
    uint64_t        numRecs;
    StratumElt_T        handOffElt;

    numRecs = memData->dataHdr.payloadSize / sizeof(PublicStratumElt_T);
    // fprintf (stderr, "processPayload of %ld records\n", numRecs);
    incomingElts = reinterpret_cast<PublicStratumElt_T*>(&memData->rawData[0]);
    for (i = 0; i < numRecs; i++)
    {

        handOffElt.Elt_ID = static_cast<SID_T>(incomingElts[i].cycleID);
        handOffElt.Elt_Trips = static_cast<STrip_T>(incomingElts[i].tripCount);

        if (handOffElt.Elt_ID != incomingElts[i].cycleID)
        {
            LogAndAbort(__FUNCTION__, "Design error - an ID exceeded 32 bits in size\n");
        }
        if (handOffElt.Elt_Trips != incomingElts[i].tripCount)
        {
            std::string ErrMsg("Consistency error in processPayload(): tripcount exceeded 32 bits - need to implement workaround\n");
            g_Logger->LogIt(ErrMsg);
        }
        ConstructStrataLayer (handOffElt);
    }
}

/// flush the partially built cycles of paths, including the cached content.
static void FlushLastCycle()
{
    SID_T   cachedTemporalID    = s_nullCycleID;
    SID_T   finalTemporalID     = s_nullCycleID;
    STrip_T cachedTemporalCount = 0;
    STrip_T finalTemporalCount  = 0;

    // Whatever partial path is there is now "complete"
    s_strataStats->IncTotalCycles(1);
    if (s_strataStats->MaxCycleLength() < s_currentStrataCycle->size())
    {
        s_strataStats->SetMaxCycleLength(s_currentStrataCycle->size());
    }
    s_strataStats->DynamicLengthHistogram()[s_currentStrataCycle->size()]++;

    // By definition, the very last path is the terminator, and thus cannot have
    // an immediate predecessor.  Thus, there will not be a match of the previously
    // cached path.  We need to process the cached content and whatever we have
    // built to date as the current
    FindOrCreatePath(&s_prevStrataCache, &cachedTemporalID, &cachedTemporalCount);
    FindOrCreatePath(s_currentStrataCycle, &finalTemporalID, &finalTemporalCount);

    PassToCompressionServer(cachedTemporalID, cachedTemporalCount, false);
    PassToCompressionServer(finalTemporalID, finalTemporalCount, true);

    s_currentStrataCycle->ResetCycle();

}

void WriteStrataDB()
{
    PureDBWrite strataInfoDB;

    strataInfoDB.Open(s_strataInfoDBName);

    // Header
    strataInfoDB.Write(s_dbVerNumKey, s_dbVerNumValue);

    // Statistics.  See dbio.h for the conventions used
    // feels like some sort of table would be more appropriate
    strataInfoDB.Write("StrataDictionary:1", s_dbStatsTotalCycles);
    strataInfoDB.Write(s_dbStatsTotalCycles, s_strataStats->TotalCycles());

    strataInfoDB.Write("StrataDictionary:2", s_dbStatsUniqueCycles);
    strataInfoDB.Write(s_dbStatsUniqueCycles, s_strataStats->UniqueCycles());

    strataInfoDB.Write("StrataDictionary:3", s_dbStatsMaxCycleLength);
    strataInfoDB.Write(s_dbStatsMaxCycleLength, s_strataStats->MaxCycleLength());

    strataInfoDB.Write("StrataDictionary:4", s_dbStatsMaxTripCount);
    strataInfoDB.Write(s_dbStatsMaxTripCount, s_strataStats->MaxTripCount());

    strataInfoDB.Write("StrataDictionary:5", s_dbStatsMaxImmediateTripCount);
    strataInfoDB.Write(s_dbStatsMaxImmediateTripCount, s_strataStats->MaxImmediateTripCount());

    strataInfoDB.Write("StrataDictionary:6", s_dbStatsCachedHits);
    strataInfoDB.Write(s_dbStatsCachedHits, s_strataStats->CachedHits());

    uint64_t    numEntries = 6;
    strataInfoDB.Write("StrataDictionary", numEntries);

    // And finally the contents of each unique strata.
    // s_nullCycleID because we increment at the end of the loop
    SID_T   sanityCheck;

    // The vectors of each strata, and net info about each
    sanityCheck = s_nullCycleID;
    for (SCycleToSInfo_T::const_iterator c_iter = s_StrataCycletoInfo->begin(); c_iter != s_StrataCycletoInfo->end(); c_iter++)
    {
        std::string     theVectorKey;
        std::string     theIDKey;
        StrataCycle_T   CycleVec    = c_iter->first;
        StratumInfo_T   thisInfo    = c_iter->second;

        theVectorKey = s_dbStratumIDKey + ToString(thisInfo.StrID);
        theIDKey = s_dbStratumInfoKey + ToString(thisInfo.StrID);

        /// Note: 6.2.3 of "The C++ Standard Library" by Josuttis notes that elements of a vector
        /// can be expected to be in contiguous memory.  Thus, we merely have to set the call
        /// to Write() up carefully, and do not need to perform a new/copy/delete.
        strataInfoDB.Write(const_cast<char*>(theVectorKey.c_str()),
                         reinterpret_cast<char*>(&CycleVec[0]),
                         CycleVec.size() * sizeof(CycleVec[0]));
        strataInfoDB.Write(theIDKey, thisInfo);
        sanityCheck++;
    }

    // Need to account for the unmapped items
    if (s_strataStats->UniqueCycles() != (sanityCheck + s_strataStats->UnmappedLongCycles()))
    {
        LogAndAbort(__FUNCTION__, "Consistency error: mismatch on number of unique paths vs. number in s_StrataCycletoInfo map\n");
    }

    strataInfoDB.Close();
}

/// Finish up, report/log relevant statistics, write content to the strata database files
void StrataFini()
{
    std::string     scribble;

    FlushLastCycle();
    if (g_saveStrata)
    {
        WriteStrataDB();
    }

    // General strata statistics
    scribble += "TotalCycles: " + ToString(s_strataStats->TotalCycles()) + "\n";
    scribble += "UniqueCycles: " + ToString(s_strataStats->UniqueCycles()) + "\n";
    scribble += "MaxCycleLength: " + ToString(s_strataStats->MaxCycleLength()) + "\n";
    scribble += "MaxTripCount: " + ToString(s_strataStats->MaxTripCount()) + "\n";
    scribble += "MaxImmediateTripCount: " + ToString(s_strataStats->MaxImmediateTripCount()) + "\n";
    scribble += "CachedHits: " + ToString(s_strataStats->CachedHits()) + "\n";
    scribble += "Unmapped Long Cycles: " + ToString(s_strataStats->UnmappedLongCycles()) + "\n";
    scribble += "Long Cycle size defined as being more than: " + ToString(s_neverMap) + "\n";
    g_Logger->LogIt(scribble);

    std::string histogramHdr;
    std::string histogramOutput;
    // Unique Path length distribution
    LengthHistogram_T   uniqueHist = s_strataStats->UniqueLengthHistogram();

    histogramHdr = "\nUnique Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = uniqueHist.begin(); c_iter != uniqueHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);

    // Dynamic path length histogram
    LengthHistogram_T   dynHist = s_strataStats->DynamicLengthHistogram();

    histogramHdr = "\nDynamic Path length distribution\n\tLength\tCount\n";
    g_Logger->LogIt(histogramHdr);

    histogramOutput = "";
    for (LengthHistogram_T::const_iterator c_iter = dynHist.begin(); c_iter != dynHist.end(); c_iter++)
    {
        histogramOutput = histogramOutput + "\t" + ToString(c_iter->first)
                                          + "\t" + ToString(c_iter->second)
                                          + "\n";
    }
    histogramOutput = histogramOutput + "\n";
    g_Logger->LogIt(histogramOutput);

    // And cleanly delete items created in StrataInit()
    delete s_StrataCycletoInfo;
    delete s_strataStats;
    delete s_currentStrataCycle;
}

// Two ways to invoke this:
//  <exe> <input mmap file> <output mmap file>
// or
//  <exe> <input dumped data file>
int main(int argc, char *argv[])
{
    int         exitcleanly = 0;
    g_argV0 = argv[0];
    g_Logger = new ProfileLogger();
    if (NULL == g_Logger)
    {
        LogAndAbort(__FUNCTION__, "Unable to set up logging facility\n");
    }

    if (2 == argc)
    {
        g_dumpMode = true;
    }
    else if (3 != argc)
    {
        return usage();
    }

    // Trusted invocation
    snprintf(g_strataMapFileName, sizeof(g_strataMapFileName) - 1, "%s", argv[1]);
    snprintf(g_compressMapFileName, sizeof(g_compressMapFileName) - 1, "%s", argv[2]);

    signal (SIGINT, cleanup_exit);

    if (g_dumpMode)
    {
        // Actual physical file that we read
        int oFlags = O_RDONLY;
        g_dumpFD = open(g_strataMapFileName, oFlags);
        if (-1 == g_dumpFD)
        {
            fprintf (stderr, "%s: Unable to open dumpfile %s\n", g_argV0, g_strataMapFileName);
            cleanup_exit(1);
        }
    }
    else
    {
        // As a server/consumer, we create the stratum mmap file, and when the
        // upstream content server passes us a "we're done" message we delete it
        unlink (g_strataMapFileName);
        s_collectorSharedMem = OpenMMapFile (argv[0], s_OpenByConsumer, g_strataMapFileName);
        if (s_collectorSharedMem == NULL) {
            cleanup_exit(1);
        }

        if (g_saveStrata)
        {
            // As a client/producer, we utilize the compression server mmap file
            // (the compression server will create and destroy it).  When we are done,
            // we pass a cleanup/shutdown message on to it
            s_compressSharedMem = OpenMMapFile (g_argV0, s_OpenByProducer, g_compressMapFileName);
            if (s_compressSharedMem == NULL) {
                fprintf (stderr, "%s: Unable to open %s as client\n", g_argV0, g_compressMapFileName);
                cleanup_exit(1);
            }
        }
    }

    s_dataPair.state = STATE_NONE;
        // And mark region as ready for use
    if (!g_dumpMode)
    {
        MarkReadyToFillData(s_collectorSharedMem);
    }

    while (!exitcleanly) {

        if (g_dumpMode)
        {
            ssize_t rdSize;

            rdSize = read(g_dumpFD, &s_incomingShM, sizeof(s_incomingShM));
            if (-1 == rdSize)
            {
                perror("Read failed");
            }
            else if (sizeof(s_incomingShM) != rdSize)
            {
                fprintf(stderr, "Problematic read - bad size read %ld, expected to read %ld\n",
                        rdSize,
                        sizeof(s_incomingShM));
            }
        }
        else
        {
            // Technique for debugging the strata server:
            //  Break on the MarkDataFilled statement in the path generator;
            //  In the strata server, set a break on the WaitUntilReadyToFill

            WaitUntilDataFilled(s_collectorSharedMem);
            // Data in the shared memory buffer, now ready to process

            // At this point, start trying to be as concurrent as possible.  Make a local
            // copy of the content, and operate on that, and then mark the shared memory
            // region as available to write new data into.  Assume that it will be almost
            // completely full, so simply copy the entire thing.  The only time it will be
            // sparse is on the final set of data.

            // Read from shared memory
            memcpy ((void*) &s_incomingShM, (void*) s_collectorSharedMem, sizeof (s_incomingShM));
            MarkReadyToFillData(s_collectorSharedMem);
        }

        // Only relevant once deployed in the field (if at all) and the format changes
        if (s_shm_verno != s_incomingShM.dataHdr.version) {
            fprintf (stderr, "%s: received data from an unsupported version client: %d\n",
                        argv[0], s_incomingShM.dataHdr.version);
            continue; 
        }

            //
        // Open operations
        // Very simple for strata analyzer: initialize strata tables.
        if (s_incomingShM.dataHdr.operation & (OP_OPEN_APPEND|OP_OPEN_TRUNC)) {
            if (s_dataPair.state == STATE_OPENED) {
                if (strcmp (s_incomingShM.dataHdr.dataFileBaseName, s_dataPair.baseFileName) == 0) {
                    fprintf (stderr, "%s: [%s] is already open... ignoring open op\n",
                                argv[0], s_incomingShM.dataHdr.dataFileBaseName);
                    goto skip_open;
                } else {
                    fprintf (stderr,
                                "%s: Unable to open (append) to [%s] - another file pair is in use: [%s]\n",
                                argv[0], s_incomingShM.dataHdr.dataFileBaseName, s_dataPair.baseFileName);
                }
                continue;
            }
            switch (s_incomingShM.dataHdr.collectionSet & s_lowNibbleMask)
            {
                case COLLECT_INSNPATHS:
                {
                    s_cSetPrefix = s_csetIPath;
                    break;
                }
                case COLLECT_DATAPATHS:
                {
                    s_cSetPrefix = s_csetDPath;
                    break;
                }
                default:
                {
                    s_cSetPrefix = "ERROR";
                    break;
                }
            }

            if (PROFILEMODE_STRATALENGTH == (s_incomingShM.dataHdr.collectionSet & s_highNibbleMask))
            {
                g_saveStrata = false;
            }

            StrataInit (s_incomingShM.dataHdr.dataFileBaseName);
            s_dataPair.state = STATE_OPENED;
        }

skip_open: ;
        // Process the data payload (if any)
        processPayload (&s_incomingShM);

        if (s_incomingShM.dataHdr.operation & (OP_FLUSH_CLOSE | OP_DELETE)) {
            StrataFini();
            s_dataPair.state = STATE_CLOSED;
            exitcleanly = 1;
        }

        // Delete not meaningful for strata_db file
        if (s_incomingShM.dataHdr.operation & OP_DELETE) {
            s_dataPair.indexFileName[0] = 0x0;
            s_dataPair.dataFileName[0] = 0x0;
            s_dataPair.state = STATE_NONE;
        }
    }

    cleanup();
    exit (0);
}
